import compileall
import os
import shutil
import subprocess
import sys
import textwrap
import venv as _venv
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

import virtualenv as _virtualenv

if TYPE_CHECKING:
    # Literal was introduced in Python 3.8.
    from typing import Literal

    VirtualEnvironmentType = Literal["virtualenv", "venv"]
else:
    VirtualEnvironmentType = str


class VirtualEnvironment:
    """
    An abstraction around virtual environments, currently it only uses
    virtualenv but in the future it could use pyvenv.
    """

    def __init__(
        self,
        location: Path,
        template: Optional["VirtualEnvironment"] = None,
        venv_type: Optional[VirtualEnvironmentType] = None,
    ):
        self.location = location
        assert template is None or venv_type is None
        self._venv_type: VirtualEnvironmentType
        if template is not None:
            self._venv_type = template._venv_type
        elif venv_type is not None:
            self._venv_type = venv_type
        else:
            self._venv_type = "virtualenv"
        self._user_site_packages = False
        self._template = template
        self._sitecustomize: Optional[str] = None
        self._update_paths()
        self._create()

    def _update_paths(self) -> None:
        home, lib, inc, bin = _virtualenv.path_locations(self.location)
        self.bin = Path(bin)
        self.site = Path(lib) / "site-packages"
        # Workaround for https://github.com/pypa/virtualenv/issues/306
        if hasattr(sys, "pypy_version_info"):
            version_dir = str(sys.version_info.major)
            self.lib = Path(home, "lib-python", version_dir)
        else:
            self.lib = Path(lib)

    def __repr__(self) -> str:
        return f"<VirtualEnvironment {self.location}>"

    def _create(self, clear: bool = False) -> None:
        if clear:
            shutil.rmtree(self.location)
        if self._template:
            # On Windows, calling `_virtualenv.path_locations(target)`
            # will have created the `target` directory...
            if sys.platform == "win32" and self.location.exists():
                self.location.rmdir()
            # Clone virtual environment from template.
            shutil.copytree(self._template.location, self.location, symlinks=True)
            self._sitecustomize = self._template.sitecustomize
            self._user_site_packages = self._template.user_site_packages
        else:
            # Create a new virtual environment.
            if self._venv_type == "virtualenv":
                subprocess.check_call(
                    [
                        sys.executable,
                        "-m",
                        "virtualenv",
                        "--no-pip",
                        "--no-wheel",
                        "--no-setuptools",
                        str(self.location),
                    ]
                )
                self._fix_virtualenv_site_module()
            elif self._venv_type == "venv":
                builder = _venv.EnvBuilder()
                context = builder.ensure_directories(self.location)
                builder.create_configuration(context)
                builder.setup_python(context)
                self.site.mkdir(parents=True, exist_ok=True)
            self.sitecustomize = self._sitecustomize
            self.user_site_packages = self._user_site_packages

    def _fix_virtualenv_site_module(self) -> None:
        # Patch `site.py` so user site work as expected.
        site_py = self.lib / "site.py"
        with open(site_py) as fp:
            site_contents = fp.read()
        for pattern, replace in (
            (
                # Ensure enabling user site does not result in adding
                # the real site-packages' directory to `sys.path`.
                ("\ndef virtual_addsitepackages(known_paths):\n"),
                (
                    "\ndef virtual_addsitepackages(known_paths):\n"
                    "    return known_paths\n"
                ),
            ),
            (
                # Fix sites ordering: user site must be added before system.
                (
                    "\n    paths_in_sys = addsitepackages(paths_in_sys)"
                    "\n    paths_in_sys = addusersitepackages(paths_in_sys)\n"
                ),
                (
                    "\n    paths_in_sys = addusersitepackages(paths_in_sys)"
                    "\n    paths_in_sys = addsitepackages(paths_in_sys)\n"
                ),
            ),
        ):
            assert pattern in site_contents
            site_contents = site_contents.replace(pattern, replace)
        with open(site_py, "w") as fp:
            fp.write(site_contents)
        # Make sure bytecode is up-to-date too.
        assert compileall.compile_file(str(site_py), quiet=1, force=True)

    def _customize_site(self) -> None:
        contents = ""
        if self._venv_type == "venv":
            # Enable user site (before system).
            contents += textwrap.dedent(
                """
                import os, site, sys

                if not os.environ.get('PYTHONNOUSERSITE', False):

                    site.ENABLE_USER_SITE = True

                    # First, drop system-sites related paths.
                    original_sys_path = sys.path[:]
                    known_paths = set()
                    for path in site.getsitepackages():
                        site.addsitedir(path, known_paths=known_paths)
                    system_paths = sys.path[len(original_sys_path):]
                    for path in system_paths:
                        if path in original_sys_path:
                            original_sys_path.remove(path)
                    sys.path = original_sys_path

                    # Second, add user-site.
                    site.addsitedir(site.getusersitepackages())

                    # Third, add back system-sites related paths.
                    for path in site.getsitepackages():
                        site.addsitedir(path)
                """
            ).strip()
        if self._sitecustomize is not None:
            contents += "\n" + self._sitecustomize
        sitecustomize = self.site / "sitecustomize.py"
        sitecustomize.write_text(contents)
        # Make sure bytecode is up-to-date too.
        assert compileall.compile_file(str(sitecustomize), quiet=1, force=True)

    def clear(self) -> None:
        self._create(clear=True)

    def move(self, location: Union[Path, str]) -> None:
        shutil.move(os.fspath(self.location), location)
        self.location = Path(location)
        self._update_paths()

    @property
    def sitecustomize(self) -> Optional[str]:
        return self._sitecustomize

    @sitecustomize.setter
    def sitecustomize(self, value: str) -> None:
        self._sitecustomize = value
        self._customize_site()

    @property
    def user_site_packages(self) -> bool:
        return self._user_site_packages

    @user_site_packages.setter
    def user_site_packages(self, value: bool) -> None:
        self._user_site_packages = value
        if self._venv_type == "virtualenv":
            marker = self.lib / "no-global-site-packages.txt"
            if self._user_site_packages:
                marker.unlink()
            else:
                marker.touch()
        elif self._venv_type == "venv":
            self._customize_site()



"""Helper classes as mocks for requests objects.
"""

from io import BytesIO
from typing import Any, Callable, Dict, Iterator, List, Optional

_Hook = Callable[["MockResponse"], None]


class FakeStream:
    def __init__(self, contents: bytes) -> None:
        self._io = BytesIO(contents)

    def read(self, size: int, decode_content: Optional[bool] = None) -> bytes:
        return self._io.read(size)

    def stream(
        self, size: int, decode_content: Optional[bool] = None
    ) -> Iterator[bytes]:
        yield self._io.read(size)

    def release_conn(self) -> None:
        pass


class MockResponse:
    request: "MockRequest"
    connection: "MockConnection"
    url: str

    def __init__(self, contents: bytes) -> None:
        self.raw = FakeStream(contents)
        self.content = contents
        self.reason = "OK"
        self.status_code = 200
        self.headers = {"Content-Length": str(len(contents))}
        self.history: List[MockResponse] = []
        self.from_cache = False


class MockConnection:
    def _send(self, req: "MockRequest", **kwargs: Any) -> MockResponse:
        raise NotImplementedError("_send must be overridden for tests")

    def send(self, req: "MockRequest", **kwargs: Any) -> MockResponse:
        resp = self._send(req, **kwargs)
        for cb in req.hooks.get("response", []):
            cb(resp)
        return resp


class MockRequest:
    def __init__(self, url: str) -> None:
        self.url = url
        self.headers: Dict[str, str] = {}
        self.hooks: Dict[str, List[_Hook]] = {}

    def register_hook(self, event_name: str, callback: _Hook) -> None:
        self.hooks.setdefault(event_name, []).append(callback)



import os

from setuptools import find_packages, setup


def path_to_url(path):
    """
    Convert a path to URI. The path will be made absolute and
    will not have quoted path parts.
    """
    path = os.path.normpath(os.path.abspath(path))
    drive, path = os.path.splitdrive(path)
    filepath = path.split(os.path.sep)
    url = "/".join(filepath)
    if drive:
        return "file:///" + drive + url
    return "file://" + url


setup(
    name="LocalExtras",
    version="0.0.2",
    packages=find_packages(),
    install_requires=["simple==1.0"],
    extras_require={"bar": ["simple==2.0"], "baz": ["singlemodule"]},
)



import os

from setuptools.build_meta import build_sdist
from setuptools.build_meta import build_wheel as setuptools_build_wheel
from setuptools.build_meta import (
    get_requires_for_build_sdist,
    get_requires_for_build_wheel,
    prepare_metadata_for_build_wheel,
)


def build_wheel(*a, **kw):
    if os.environ.get("PIP_TEST_FAIL_BUILD_WHEEL"):
        raise RuntimeError("Failing build_wheel, as requested.")

    # Create the marker file to record that the hook was called
    with open(os.environ["PIP_TEST_MARKER_FILE"], "wb"):
        pass

    return setuptools_build_wheel(*a, **kw)



"""Tests for the ``pip hash`` command"""
from pathlib import Path

from tests.lib import PipTestEnvironment


def test_basic_hash(script: PipTestEnvironment, tmpdir: Path) -> None:
    """Run 'pip hash' through its default behavior."""
    expected = (
        "--hash=sha256:2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425"
        "e73043362938b9824"
    )
    result = script.pip("hash", _hello_file(tmpdir))
    assert expected in str(result)


def test_good_algo_option(script: PipTestEnvironment, tmpdir: Path) -> None:
    """Make sure the -a option works."""
    expected = (
        "--hash=sha512:9b71d224bd62f3785d96d46ad3ea3d73319bfbc2890caad"
        "ae2dff72519673ca72323c3d99ba5c11d7c7acc6e14b8c5da0c4663475c2e"
        "5c3adef46f73bcdec043"
    )
    result = script.pip("hash", "-a", "sha512", _hello_file(tmpdir))
    assert expected in str(result)


def test_bad_algo_option(script: PipTestEnvironment, tmpdir: Path) -> None:
    """Make sure the -a option raises an error when given a bad operand."""
    result = script.pip(
        "hash", "-a", "invalidname", _hello_file(tmpdir), expect_error=True
    )
    assert "invalid choice: 'invalidname'" in str(result)


def _hello_file(tmpdir: Path) -> Path:
    """Return a temp file to hash containing "hello"."""
    file = tmpdir / "hashable"
    file.write_text("hello")
    return file



"""Basic CLI functionality checks.
"""
from textwrap import dedent

import pytest

from tests.lib import PipTestEnvironment


@pytest.mark.parametrize(
    "entrypoint",
    [
        ("fake_pip = pip._internal.main:main",),
        ("fake_pip = pip._internal:main",),
        ("fake_pip = pip:main",),
    ],
)
def test_entrypoints_work(entrypoint: str, script: PipTestEnvironment) -> None:
    if script.zipapp:
        pytest.skip("Zipapp does not include entrypoints")

    fake_pkg = script.temp_path / "fake_pkg"
    fake_pkg.mkdir()
    fake_pkg.joinpath("setup.py").write_text(
        dedent(
            """
    from setuptools import setup

    setup(
        name="fake-pip",
        version="0.1.0",
        entry_points={{
            "console_scripts": [
                {!r}
            ]
        }}
    )
    """.format(
                entrypoint
            )
        )
    )

    # expect_temp because pip install will generate fake_pkg.egg-info
    script.pip("install", "-vvv", str(fake_pkg), expect_temp=True)
    result = script.pip("-V")
    result2 = script.run("fake_pip", "-V", allow_stderr_warning=True)
    assert result.stdout == result2.stdout
    assert "old script wrapper" in result2.stderr



import os
import re
from pathlib import Path
from typing import Optional

from pip._internal.models.direct_url import DIRECT_URL_METADATA_NAME, DirectUrl
from tests.lib import TestPipResult


def get_created_direct_url_path(result: TestPipResult, pkg: str) -> Optional[Path]:
    direct_url_metadata_re = re.compile(
        pkg + r"-[\d\.]+\.dist-info." + DIRECT_URL_METADATA_NAME + r"$"
    )
    for filename in result.files_created:
        if direct_url_metadata_re.search(os.fspath(filename)):
            return result.test_env.base_path / filename
    return None


def get_created_direct_url(result: TestPipResult, pkg: str) -> Optional[DirectUrl]:
    direct_url_path = get_created_direct_url_path(result, pkg)
    if direct_url_path:
        with open(direct_url_path) as f:
            return DirectUrl.from_json(f.read())
    return None



import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Tuple, Union

import pytest

from tests.conftest import ScriptFactory
from tests.lib import PipTestEnvironment, TestData, TestPipResult

if TYPE_CHECKING:
    from typing import Protocol
else:
    # TODO: Protocol was introduced in Python 3.8. Remove this branch when
    # dropping support for Python 3.7.
    Protocol = object


COMPLETION_FOR_SUPPORTED_SHELLS_TESTS = (
    (
        "bash",
        """\
_pip_completion()
{
    COMPREPLY=( $( COMP_WORDS="${COMP_WORDS[*]}" \\
                   COMP_CWORD=$COMP_CWORD \\
                   PIP_AUTO_COMPLETE=1 $1 2>/dev/null ) )
}
complete -o default -F _pip_completion pip""",
    ),
    (
        "fish",
        """\
function __fish_complete_pip
    set -lx COMP_WORDS (commandline -o) ""
    set -lx COMP_CWORD ( \\
        math (contains -i -- (commandline -t) $COMP_WORDS)-1 \\
    )
    set -lx PIP_AUTO_COMPLETE 1
    string split \\  -- (eval $COMP_WORDS[1])
end
complete -fa "(__fish_complete_pip)" -c pip""",
    ),
    (
        "zsh",
        """\
function _pip_completion {
  local words cword
  read -Ac words
  read -cn cword
  reply=( $( COMP_WORDS="$words[*]" \\
             COMP_CWORD=$(( cword-1 )) \\
             PIP_AUTO_COMPLETE=1 $words[1] 2>/dev/null ))
}
compctl -K _pip_completion pip""",
    ),
    (
        "powershell",
        """\
if ((Test-Path Function:\\TabExpansion) -and -not `
    (Test-Path Function:\\_pip_completeBackup)) {
    Rename-Item Function:\\TabExpansion _pip_completeBackup
}
function TabExpansion($line, $lastWord) {
    $lastBlock = [regex]::Split($line, '[|;]')[-1].TrimStart()
    if ($lastBlock.StartsWith("pip ")) {
        $Env:COMP_WORDS=$lastBlock
        $Env:COMP_CWORD=$lastBlock.Split().Length - 1
        $Env:PIP_AUTO_COMPLETE=1
        (& pip).Split()
        Remove-Item Env:COMP_WORDS
        Remove-Item Env:COMP_CWORD
        Remove-Item Env:PIP_AUTO_COMPLETE
    }
    elseif (Test-Path Function:\\_pip_completeBackup) {
        # Fall back on existing tab expansion
        _pip_completeBackup $line $lastWord
    }
}""",
    ),
)


@pytest.fixture(scope="session")
def script_with_launchers(
    tmpdir_factory: pytest.TempPathFactory,
    script_factory: ScriptFactory,
    common_wheels: Path,
    pip_src: Path,
) -> PipTestEnvironment:
    tmpdir = tmpdir_factory.mktemp("script_with_launchers")
    script = script_factory(tmpdir.joinpath("workspace"))
    # Re-install pip so we get the launchers.
    script.pip_install_local("-f", common_wheels, pip_src)
    return script


@pytest.mark.parametrize(
    "shell, completion",
    COMPLETION_FOR_SUPPORTED_SHELLS_TESTS,
    ids=[t[0] for t in COMPLETION_FOR_SUPPORTED_SHELLS_TESTS],
)
def test_completion_for_supported_shells(
    script_with_launchers: PipTestEnvironment, shell: str, completion: str
) -> None:
    """
    Test getting completion for bash shell
    """
    result = script_with_launchers.pip("completion", "--" + shell, use_module=False)
    actual = str(result.stdout)
    if script_with_launchers.zipapp:
        # The zipapp reports its name as "pip.pyz", but the expected
        # output assumes "pip"
        actual = actual.replace("pip.pyz", "pip")
    assert completion in actual, actual


@pytest.fixture(scope="session")
def autocomplete_script(
    tmpdir_factory: pytest.TempPathFactory, script_factory: ScriptFactory
) -> PipTestEnvironment:
    tmpdir = tmpdir_factory.mktemp("autocomplete_script")
    return script_factory(tmpdir.joinpath("workspace"))


class DoAutocomplete(Protocol):
    def __call__(
        self, words: str, cword: str, cwd: Union[Path, str, None] = None
    ) -> Tuple[TestPipResult, PipTestEnvironment]:
        ...


@pytest.fixture
def autocomplete(
    autocomplete_script: PipTestEnvironment, monkeypatch: pytest.MonkeyPatch
) -> DoAutocomplete:
    monkeypatch.setattr(autocomplete_script, "environ", os.environ.copy())
    autocomplete_script.environ["PIP_AUTO_COMPLETE"] = "1"

    def do_autocomplete(
        words: str, cword: str, cwd: Union[Path, str, None] = None
    ) -> Tuple[TestPipResult, PipTestEnvironment]:
        autocomplete_script.environ["COMP_WORDS"] = words
        autocomplete_script.environ["COMP_CWORD"] = cword
        result = autocomplete_script.run(
            "python",
            "-c",
            "from pip._internal.cli.autocompletion import autocomplete;"
            "autocomplete()",
            expect_error=True,
            cwd=cwd,
        )

        return result, autocomplete_script

    return do_autocomplete


def test_completion_for_unknown_shell(autocomplete_script: PipTestEnvironment) -> None:
    """
    Test getting completion for an unknown shell
    """
    error_msg = "no such option: --myfooshell"
    result = autocomplete_script.pip("completion", "--myfooshell", expect_error=True)
    assert error_msg in result.stderr, "tests for an unknown shell failed"


def test_completion_alone(autocomplete_script: PipTestEnvironment) -> None:
    """
    Test getting completion for none shell, just pip completion
    """
    result = autocomplete_script.pip("completion", allow_stderr_error=True)
    assert (
        "ERROR: You must pass --bash or --fish or --powershell or --zsh"
        in result.stderr
    ), ("completion alone failed -- " + result.stderr)


def test_completion_for_un_snippet(autocomplete: DoAutocomplete) -> None:
    """
    Test getting completion for ``un`` should return uninstall
    """

    res, env = autocomplete("pip un", "1")
    assert res.stdout.strip().split() == ["uninstall"], res.stdout


def test_completion_for_default_parameters(autocomplete: DoAutocomplete) -> None:
    """
    Test getting completion for ``--`` should contain --help
    """

    res, env = autocomplete("pip --", "1")
    assert "--help" in res.stdout, "autocomplete function could not complete ``--``"


def test_completion_option_for_command(autocomplete: DoAutocomplete) -> None:
    """
    Test getting completion for ``--`` in command (e.g. ``pip search --``)
    """

    res, env = autocomplete("pip search --", "2")
    assert "--help" in res.stdout, "autocomplete function could not complete ``--``"


def test_completion_short_option(autocomplete: DoAutocomplete) -> None:
    """
    Test getting completion for short options after ``-`` (eg. pip -)
    """

    res, env = autocomplete("pip -", "1")

    assert (
        "-h" in res.stdout.split()
    ), "autocomplete function could not complete short options after ``-``"


def test_completion_short_option_for_command(autocomplete: DoAutocomplete) -> None:
    """
    Test getting completion for short options after ``-`` in command
    (eg. pip search -)
    """

    res, env = autocomplete("pip search -", "2")

    assert (
        "-h" in res.stdout.split()
    ), "autocomplete function could not complete short options after ``-``"


def test_completion_files_after_option(
    autocomplete: DoAutocomplete, data: TestData
) -> None:
    """
    Test getting completion for <file> or <dir> after options in command
    (e.g. ``pip install -r``)
    """
    res, env = autocomplete(
        words=("pip install -r r"),
        cword="3",
        cwd=data.completion_paths,
    )
    assert (
        "requirements.txt" in res.stdout
    ), "autocomplete function could not complete <file> after options in command"
    assert (
        os.path.join("resources", "") in res.stdout
    ), "autocomplete function could not complete <dir> after options in command"
    assert not any(
        out in res.stdout for out in (os.path.join("REPLAY", ""), "README.txt")
    ), (
        "autocomplete function completed <file> or <dir> that "
        "should not be completed"
    )
    if sys.platform != "win32":
        return
    assert (
        "readme.txt" in res.stdout
    ), "autocomplete function could not complete <file> after options in command"
    assert (
        os.path.join("replay", "") in res.stdout
    ), "autocomplete function could not complete <dir> after options in command"


def test_completion_not_files_after_option(
    autocomplete: DoAutocomplete, data: TestData
) -> None:
    """
    Test not getting completion files after options which not applicable
    (e.g. ``pip wheel``)
    """
    res, env = autocomplete(
        words=("pip wheel r"),
        cword="2",
        cwd=data.completion_paths,
    )
    assert not any(
        out in res.stdout
        for out in (
            "requirements.txt",
            "readme.txt",
        )
    ), "autocomplete function completed <file> when it should not complete"
    assert not any(
        os.path.join(out, "") in res.stdout for out in ("replay", "resources")
    ), "autocomplete function completed <dir> when it should not complete"


def test_pip_install_complete_files(
    autocomplete: DoAutocomplete, data: TestData
) -> None:
    """``pip install`` autocompletes wheel and sdist files."""
    res, env = autocomplete(
        words=("pip install r"),
        cword="2",
        cwd=data.completion_paths,
    )
    assert all(
        out in res.stdout
        for out in (
            "requirements.txt",
            "resources",
        )
    ), "autocomplete function could not complete <path>"


@pytest.mark.parametrize("cl_opts", ["-U", "--user", "-h"])
def test_completion_not_files_after_nonexpecting_option(
    autocomplete: DoAutocomplete, data: TestData, cl_opts: str
) -> None:
    """
    Test not getting completion files after options which not applicable
    (e.g. ``pip install``)
    """
    res, env = autocomplete(
        words=(f"pip install {cl_opts} r"),
        cword="2",
        cwd=data.completion_paths,
    )
    assert not any(
        out in res.stdout
        for out in (
            "requirements.txt",
            "readme.txt",
        )
    ), "autocomplete function completed <file> when it should not complete"
    assert not any(
        os.path.join(out, "") in res.stdout for out in ("replay", "resources")
    ), "autocomplete function completed <dir> when it should not complete"


def test_completion_directories_after_option(
    autocomplete: DoAutocomplete, data: TestData
) -> None:
    """
    Test getting completion <dir> after options in command
    (e.g. ``pip --cache-dir``)
    """
    res, env = autocomplete(
        words=("pip --cache-dir r"),
        cword="2",
        cwd=data.completion_paths,
    )
    assert (
        os.path.join("resources", "") in res.stdout
    ), "autocomplete function could not complete <dir> after options"
    assert not any(
        out in res.stdout
        for out in ("requirements.txt", "README.txt", os.path.join("REPLAY", ""))
    ), "autocomplete function completed <dir> when it should not complete"
    if sys.platform == "win32":
        assert (
            os.path.join("replay", "") in res.stdout
        ), "autocomplete function could not complete <dir> after options"


def test_completion_subdirectories_after_option(
    autocomplete: DoAutocomplete, data: TestData
) -> None:
    """
    Test getting completion <dir> after options in command
    given path of a directory
    """
    res, env = autocomplete(
        words=("pip --cache-dir " + os.path.join("resources", "")),
        cword="2",
        cwd=data.completion_paths,
    )
    assert os.path.join("resources", os.path.join("images", "")) in res.stdout, (
        "autocomplete function could not complete <dir> "
        "given path of a directory after options"
    )


def test_completion_path_after_option(
    autocomplete: DoAutocomplete, data: TestData
) -> None:
    """
    Test getting completion <path> after options in command
    given absolute path
    """
    res, env = autocomplete(
        words=("pip install -e " + os.path.join(data.completion_paths, "R")),
        cword="3",
    )
    assert all(
        os.path.normcase(os.path.join(data.completion_paths, out)) in res.stdout
        for out in ("README.txt", os.path.join("REPLAY", ""))
    ), (
        "autocomplete function could not complete <path> "
        "after options in command given absolute path"
    )


@pytest.mark.parametrize("flag", ["--bash", "--zsh", "--fish", "--powershell"])
def test_completion_uses_same_executable_name(
    autocomplete_script: PipTestEnvironment, flag: str, deprecated_python: bool
) -> None:
    executable_name = "pip{}".format(sys.version_info[0])
    # Deprecated python versions produce an extra deprecation warning
    result = autocomplete_script.run(
        executable_name,
        "completion",
        flag,
        expect_stderr=deprecated_python,
    )
    assert executable_name in result.stdout



"""
Tests for compatibility workarounds.

"""
import os

import pytest

from tests.lib import pyversion  # noqa: F401
from tests.lib import PipTestEnvironment, TestData, assert_all_changes


@pytest.mark.network
def test_debian_egg_name_workaround(script: PipTestEnvironment) -> None:
    """
    We can uninstall packages installed with the pyversion removed from the
    egg-info metadata directory name.

    Refs:
    http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=618367
    https://bugs.launchpad.net/ubuntu/+source/distribute/+bug/725178
    https://bitbucket.org/ianb/pip/issue/104/pip-uninstall-on-ubuntu-linux

    """
    result = script.pip("install", "INITools==0.2")

    egg_info = os.path.join(
        script.site_packages, f"INITools-0.2-py{pyversion}.egg-info"
    )

    # Debian only removes pyversion for global installs, not inside a venv
    # so even if this test runs on a Debian/Ubuntu system with broken
    # setuptools, since our test runs inside a venv we'll still have the normal
    # .egg-info
    result.did_create(egg_info, message=f"Couldn't find {egg_info}")

    # The Debian no-pyversion version of the .egg-info
    mangled = os.path.join(script.site_packages, "INITools-0.2.egg-info")
    result.did_not_create(mangled, message=f"Found unexpected {mangled}")

    # Simulate a Debian install by copying the .egg-info to their name for it
    full_egg_info = os.path.join(script.base_path, egg_info)
    assert os.path.isdir(full_egg_info)
    full_mangled = os.path.join(script.base_path, mangled)
    os.renames(full_egg_info, full_mangled)
    assert os.path.isdir(full_mangled)

    # Try the uninstall and verify that everything is removed.
    result2 = script.pip("uninstall", "INITools", "-y")
    assert_all_changes(result, result2, [script.venv / "build", "cache"])


def test_setup_py_with_dos_line_endings(
    script: PipTestEnvironment, data: TestData
) -> None:
    """
    It doesn't choke on a setup.py file that uses DOS line endings (\\r\\n).

    Refs https://github.com/pypa/pip/issues/237
    """
    to_install = data.packages.joinpath("LineEndings")
    script.pip("install", to_install)



import os
from functools import partial
from pathlib import Path
from unittest import mock

from pip._internal.models.direct_url import ArchiveInfo, DirectUrl, DirInfo, VcsInfo
from pip._internal.models.link import Link
from pip._internal.utils.direct_url_helpers import (
    direct_url_as_pep440_direct_reference,
    direct_url_from_link,
)
from pip._internal.vcs.git import Git


def test_as_pep440_requirement_archive() -> None:
    direct_url = DirectUrl(
        url="file:///home/user/archive.tgz",
        info=ArchiveInfo(),
    )
    direct_url.validate()
    assert (
        direct_url_as_pep440_direct_reference(direct_url, "pkg")
        == "pkg @ file:///home/user/archive.tgz"
    )
    direct_url.subdirectory = "subdir"
    direct_url.validate()
    assert (
        direct_url_as_pep440_direct_reference(direct_url, "pkg")
        == "pkg @ file:///home/user/archive.tgz#subdirectory=subdir"
    )
    assert isinstance(direct_url.info, ArchiveInfo)
    direct_url.info.hash = "sha1=1b8c5bc61a86f377fea47b4276c8c8a5842d2220"
    direct_url.validate()
    assert (
        direct_url_as_pep440_direct_reference(direct_url, "pkg")
        == "pkg @ file:///home/user/archive.tgz"
        "#sha1=1b8c5bc61a86f377fea47b4276c8c8a5842d2220&subdirectory=subdir"
    )


def test_as_pep440_requirement_dir() -> None:
    direct_url = DirectUrl(
        url="file:///home/user/project",
        info=DirInfo(editable=False),
    )
    direct_url.validate()
    assert (
        direct_url_as_pep440_direct_reference(direct_url, "pkg")
        == "pkg @ file:///home/user/project"
    )


def test_as_pep440_requirement_editable_dir() -> None:
    # direct_url_as_pep440_direct_reference behaves the same
    # irrespective of the editable flag. It's the responsibility of
    # callers to render it as editable
    direct_url = DirectUrl(
        url="file:///home/user/project",
        info=DirInfo(editable=True),
    )
    direct_url.validate()
    assert (
        direct_url_as_pep440_direct_reference(direct_url, "pkg")
        == "pkg @ file:///home/user/project"
    )


def test_as_pep440_requirement_vcs() -> None:
    direct_url = DirectUrl(
        url="https:///g.c/u/p.git",
        info=VcsInfo(vcs="git", commit_id="1b8c5bc61a86f377fea47b4276c8c8a5842d2220"),
    )
    direct_url.validate()
    assert (
        direct_url_as_pep440_direct_reference(direct_url, "pkg")
        == "pkg @ git+https:///g.c/u/p.git"
        "@1b8c5bc61a86f377fea47b4276c8c8a5842d2220"
    )
    direct_url.subdirectory = "subdir"
    direct_url.validate()
    assert (
        direct_url_as_pep440_direct_reference(direct_url, "pkg")
        == "pkg @ git+https:///g.c/u/p.git"
        "@1b8c5bc61a86f377fea47b4276c8c8a5842d2220#subdirectory=subdir"
    )


@mock.patch("pip._internal.vcs.git.Git.get_revision")
def test_from_link_vcs(mock_get_backend_for_scheme: mock.Mock) -> None:
    _direct_url_from_link = partial(direct_url_from_link, source_dir="...")
    direct_url = _direct_url_from_link(Link("git+https://g.c/u/p.git"))
    assert direct_url.url == "https://g.c/u/p.git"
    assert isinstance(direct_url.info, VcsInfo)
    assert direct_url.info.vcs == "git"
    direct_url = _direct_url_from_link(Link("git+https://g.c/u/p.git#egg=pkg"))
    assert direct_url.url == "https://g.c/u/p.git"
    direct_url = _direct_url_from_link(
        Link("git+https://g.c/u/p.git#egg=pkg&subdirectory=subdir")
    )
    assert direct_url.url == "https://g.c/u/p.git"
    assert direct_url.subdirectory == "subdir"
    direct_url = _direct_url_from_link(Link("git+https://g.c/u/p.git@branch"))
    assert direct_url.url == "https://g.c/u/p.git"
    assert isinstance(direct_url.info, VcsInfo)
    assert direct_url.info.requested_revision == "branch"
    direct_url = _direct_url_from_link(Link("git+https://g.c/u/p.git@branch#egg=pkg"))
    assert direct_url.url == "https://g.c/u/p.git"
    assert isinstance(direct_url.info, VcsInfo)
    assert direct_url.info.requested_revision == "branch"
    direct_url = _direct_url_from_link(Link("git+https://token@g.c/u/p.git"))
    assert direct_url.to_dict()["url"] == "https://g.c/u/p.git"


def test_from_link_vcs_with_source_dir_obtains_commit_id(tmpdir: Path) -> None:
    repo_path = tmpdir / "test-repo"
    repo_path.mkdir()
    repo_dir = os.fspath(repo_path)
    Git.run_command(["init"], cwd=repo_dir)
    (repo_path / "somefile").touch()
    Git.run_command(["add", "."], cwd=repo_dir)
    Git.run_command(["commit", "-m", "commit msg"], cwd=repo_dir)
    commit_id = Git.get_revision(repo_dir)
    direct_url = direct_url_from_link(
        Link("git+https://g.c/u/p.git"), source_dir=repo_dir
    )
    assert direct_url.url == "https://g.c/u/p.git"
    assert isinstance(direct_url.info, VcsInfo)
    assert direct_url.info.commit_id == commit_id


def test_from_link_vcs_without_source_dir() -> None:
    direct_url = direct_url_from_link(
        Link("git+https://g.c/u/p.git@1"), link_is_in_wheel_cache=True
    )
    assert direct_url.url == "https://g.c/u/p.git"
    assert isinstance(direct_url.info, VcsInfo)
    assert direct_url.info.commit_id == "1"


def test_from_link_archive() -> None:
    direct_url = direct_url_from_link(Link("https://g.c/archive.tgz"))
    assert direct_url.url == "https://g.c/archive.tgz"
    assert isinstance(direct_url.info, ArchiveInfo)
    direct_url = direct_url_from_link(
        Link("https://g.c/archive.tgz#sha1=1b8c5bc61a86f377fea47b4276c8c8a5842d2220")
    )
    assert isinstance(direct_url.info, ArchiveInfo)
    assert direct_url.info.hash == "sha1=1b8c5bc61a86f377fea47b4276c8c8a5842d2220"


def test_from_link_dir(tmpdir: Path) -> None:
    dir_url = tmpdir.as_uri()
    direct_url = direct_url_from_link(Link(dir_url))
    assert direct_url.url == dir_url
    assert isinstance(direct_url.info, DirInfo)


def test_from_link_hide_user_password() -> None:
    # Basic test only here, other variants are covered by
    # direct_url.redact_url tests.
    direct_url = direct_url_from_link(
        Link("git+https://user:password@g.c/u/p.git@branch#egg=pkg"),
        link_is_in_wheel_cache=True,
    )
    assert direct_url.to_dict()["url"] == "https://g.c/u/p.git"
    direct_url = direct_url_from_link(
        Link("git+ssh://git@g.c/u/p.git@branch#egg=pkg"),
        link_is_in_wheel_cache=True,
    )
    assert direct_url.to_dict()["url"] == "ssh://git@g.c/u/p.git"



import sys
from typing import Any, Callable

import pytest

from tests.lib import PipTestEnvironment, TestPipResult

PipRunner = Callable[..., TestPipResult]


@pytest.fixture()
def pip(script: PipTestEnvironment) -> PipRunner:
    def pip(*args: str, **kwargs: Any) -> TestPipResult:
        return script.pip(*args, "--use-feature=truststore", **kwargs)

    return pip


@pytest.mark.skipif(sys.version_info >= (3, 10), reason="3.10 can run truststore")
def test_truststore_error_on_old_python(pip: PipRunner) -> None:
    result = pip(
        "install",
        "--no-index",
        "does-not-matter",
        expect_error=True,
    )
    assert "The truststore feature is only available for Python 3.10+" in result.stderr


@pytest.mark.skipif(sys.version_info < (3, 10), reason="3.10+ required for truststore")
def test_truststore_error_without_preinstalled(pip: PipRunner) -> None:
    result = pip(
        "install",
        "--no-index",
        "does-not-matter",
        expect_error=True,
    )
    assert (
        "To use the truststore feature, 'truststore' must be installed into "
        "pip's current environment."
    ) in result.stderr


@pytest.mark.skipif(sys.version_info < (3, 10), reason="3.10+ required for truststore")
@pytest.mark.network
@pytest.mark.parametrize(
    "package",
    [
        "INITools",
        "https://github.com/pypa/pip-test-package/archive/refs/heads/master.zip",
    ],
    ids=["PyPI", "GitHub"],
)
def test_trustore_can_install(
    script: PipTestEnvironment,
    pip: PipRunner,
    package: str,
) -> None:
    script.pip("install", "truststore")
    result = pip("install", package)
    assert "Successfully installed" in result.stdout



from typing import Iterable

from tests.lib import PipTestEnvironment, create_test_package_with_setup


def assert_contains_expected_lines(string: str, expected_lines: Iterable[str]) -> None:
    for expected_line in expected_lines:
        assert (expected_line + "\n") in string


def test_check_install_canonicalization(script: PipTestEnvironment) -> None:
    pkga_path = create_test_package_with_setup(
        script,
        name="pkgA",
        version="1.0",
        install_requires=["normal-missing", "SPECIAL.missing"],
    )
    normal_path = create_test_package_with_setup(
        script,
        name="normal-missing",
        version="0.1",
    )
    special_path = create_test_package_with_setup(
        script,
        name="SPECIAL.missing",
        version="0.1",
    )

    # Let's install pkgA without its dependency
    result = script.pip("install", "--no-index", pkga_path, "--no-deps")
    assert "Successfully installed pkgA-1.0" in result.stdout, str(result)

    # Install the first missing dependency. Only an error for the
    # second dependency should remain.
    result = script.pip(
        "install",
        "--no-index",
        normal_path,
        "--quiet",
        allow_stderr_error=True,
    )
    expected_lines = [
        "pkga 1.0 requires SPECIAL.missing, which is not installed.",
    ]
    # Deprecated python versions produce an extra warning on stderr
    assert_contains_expected_lines(result.stderr, expected_lines)
    assert result.returncode == 0

    # Install the second missing package and expect that there is no warning
    # during the installation. This is special as the package name requires
    # name normalization (as in https://github.com/pypa/pip/issues/5134)
    result = script.pip(
        "install",
        "--no-index",
        special_path,
        "--quiet",
    )
    assert "requires" not in result.stderr
    assert result.returncode == 0

    # Double check that all errors are resolved in the end
    result = script.pip("check")
    expected_lines = [
        "No broken requirements found.",
    ]
    assert_contains_expected_lines(result.stdout, expected_lines)
    assert result.returncode == 0


def test_check_install_does_not_warn_for_out_of_graph_issues(
    script: PipTestEnvironment,
) -> None:
    pkg_broken_path = create_test_package_with_setup(
        script,
        name="broken",
        version="1.0",
        install_requires=["missing", "conflict < 1.0"],
    )
    pkg_unrelated_path = create_test_package_with_setup(
        script,
        name="unrelated",
        version="1.0",
    )
    pkg_conflict_path = create_test_package_with_setup(
        script,
        name="conflict",
        version="1.0",
    )

    # Install a package without it's dependencies
    result = script.pip("install", "--no-index", pkg_broken_path, "--no-deps")
    assert "requires" not in result.stderr

    # Install conflict package
    result = script.pip(
        "install",
        "--no-index",
        pkg_conflict_path,
        allow_stderr_error=True,
    )
    assert_contains_expected_lines(
        result.stderr,
        [
            "broken 1.0 requires missing, which is not installed.",
            "broken 1.0 requires conflict<1.0, "
            "but you have conflict 1.0 which is incompatible.",
        ],
    )

    # Install unrelated package
    result = script.pip(
        "install",
        "--no-index",
        pkg_unrelated_path,
        "--quiet",
    )
    # should not warn about broken's deps when installing unrelated package
    assert "requires" not in result.stderr

    result = script.pip("check", expect_error=True)
    expected_lines = [
        "broken 1.0 requires missing, which is not installed.",
        "broken 1.0 has requirement conflict<1.0, but you have conflict 1.0.",
    ]
    assert_contains_expected_lines(result.stdout, expected_lines)



import base64
import csv
import distutils
import hashlib
import os
import shutil
from pathlib import Path
from typing import Any

import pytest

from tests.lib import PipTestEnvironment, TestData, create_basic_wheel_for_package
from tests.lib.wheel import WheelBuilder, make_wheel


# assert_installed expects a package subdirectory, so give it to them
def make_wheel_with_file(name: str, version: str, **kwargs: Any) -> WheelBuilder:
    extra_files = kwargs.setdefault("extra_files", {})
    extra_files[f"{name}/__init__.py"] = "# example"
    return make_wheel(name=name, version=version, **kwargs)


def test_install_from_future_wheel_version(
    script: PipTestEnvironment, tmpdir: Path
) -> None:
    """
    Test installing a wheel with a WHEEL metadata version that is:
    - a major version ahead of what we expect (not ok), and
    - a minor version ahead of what we expect (ok)
    """
    from tests.lib import TestFailure

    package = make_wheel_with_file(
        name="futurewheel",
        version="3.0",
        wheel_metadata_updates={"Wheel-Version": "3.0"},
    ).save_to_dir(tmpdir)

    result = script.pip("install", package, "--no-index", expect_error=True)
    with pytest.raises(TestFailure):
        result.assert_installed("futurewheel", without_egg_link=True, editable=False)

    package = make_wheel_with_file(
        name="futurewheel",
        version="1.9",
        wheel_metadata_updates={"Wheel-Version": "1.9"},
    ).save_to_dir(tmpdir)
    result = script.pip("install", package, "--no-index", expect_stderr=True)
    result.assert_installed("futurewheel", without_egg_link=True, editable=False)


@pytest.mark.parametrize(
    "wheel_name",
    [
        "brokenwheel-1.0-py2.py3-none-any.whl",
        "corruptwheel-1.0-py2.py3-none-any.whl",
    ],
)
def test_install_from_broken_wheel(
    script: PipTestEnvironment, data: TestData, wheel_name: str
) -> None:
    """
    Test that installing a broken wheel fails properly
    """
    from tests.lib import TestFailure

    package = data.packages.joinpath(wheel_name)
    result = script.pip("install", package, "--no-index", expect_error=True)
    with pytest.raises(TestFailure):
        result.assert_installed("futurewheel", without_egg_link=True, editable=False)


def test_basic_install_from_wheel(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing from a wheel (that has a script)
    """
    shutil.copy(shared_data.packages / "has.script-1.0-py2.py3-none-any.whl", tmpdir)
    result = script.pip(
        "install",
        "has.script==1.0",
        "--no-index",
        "--find-links",
        tmpdir,
    )
    dist_info_folder = script.site_packages / "has.script-1.0.dist-info"
    result.did_create(dist_info_folder)
    script_file = script.bin / "script.py"
    result.did_create(script_file)


def test_basic_install_from_wheel_with_extras(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing from a wheel with extras.
    """
    shutil.copy(shared_data.packages / "complex_dist-0.1-py2.py3-none-any.whl", tmpdir)
    shutil.copy(shared_data.packages / "simple.dist-0.1-py2.py3-none-any.whl", tmpdir)
    result = script.pip(
        "install",
        "complex-dist[simple]",
        "--no-index",
        "--find-links",
        tmpdir,
    )
    dist_info_folder = script.site_packages / "complex_dist-0.1.dist-info"
    result.did_create(dist_info_folder)
    dist_info_folder = script.site_packages / "simple.dist-0.1.dist-info"
    result.did_create(dist_info_folder)


def test_basic_install_from_wheel_file(
    script: PipTestEnvironment, data: TestData
) -> None:
    """
    Test installing directly from a wheel file.
    """
    package = data.packages.joinpath("simple.dist-0.1-py2.py3-none-any.whl")
    result = script.pip("install", package, "--no-index")
    dist_info_folder = script.site_packages / "simple.dist-0.1.dist-info"
    result.did_create(dist_info_folder)
    installer = dist_info_folder / "INSTALLER"
    result.did_create(installer)
    with open(script.base_path / installer, "rb") as installer_file:
        installer_details = installer_file.read()
        assert installer_details == b"pip\n"
    installer_temp = dist_info_folder / "INSTALLER.pip"
    result.did_not_create(installer_temp)


# Installation seems to work, but scripttest fails to check.
# I really don't care now since we're desupporting it soon anyway.
def test_basic_install_from_unicode_wheel(
    script: PipTestEnvironment, data: TestData
) -> None:
    """
    Test installing from a wheel (that has a script)
    """
    make_wheel(
        "unicode_package",
        "1.0",
        extra_files={
            "வணக்கம்/__init__.py": b"",
            "வணக்கம்/નમસ્તે.py": b"",
        },
    ).save_to_dir(script.scratch_path)

    result = script.pip(
        "install",
        "unicode_package==1.0",
        "--no-index",
        "--find-links",
        script.scratch_path,
    )
    dist_info_folder = script.site_packages / "unicode_package-1.0.dist-info"
    result.did_create(dist_info_folder)

    file1 = script.site_packages.joinpath("வணக்கம்", "__init__.py")
    result.did_create(file1)

    file2 = script.site_packages.joinpath("வணக்கம்", "નમસ્તે.py")
    result.did_create(file2)


def get_header_scheme_path_for_script(
    script: PipTestEnvironment, dist_name: str
) -> Path:
    command = (
        "from pip._internal.locations import get_scheme;"
        "scheme = get_scheme({!r});"
        "print(scheme.headers);"
    ).format(dist_name)
    result = script.run("python", "-c", command).stdout
    return Path(result.strip())


def test_install_from_wheel_with_headers(script: PipTestEnvironment) -> None:
    """
    Test installing from a wheel file with headers
    """
    header_text = "/* hello world */\n"
    package = make_wheel(
        "headers.dist",
        "0.1",
        extra_data_files={"headers/header.h": header_text},
    ).save_to_dir(script.scratch_path)
    result = script.pip("install", package, "--no-index")
    dist_info_folder = script.site_packages / "headers.dist-0.1.dist-info"
    result.did_create(dist_info_folder)

    header_scheme_path = get_header_scheme_path_for_script(script, "headers.dist")
    header_path = header_scheme_path / "header.h"
    assert header_path.read_text() == header_text


@pytest.mark.usefixtures("with_wheel")
def test_install_wheel_with_target(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing a wheel using pip install --target
    """
    shutil.copy(shared_data.packages / "simple.dist-0.1-py2.py3-none-any.whl", tmpdir)
    target_dir = script.scratch_path / "target"
    result = script.pip(
        "install",
        "simple.dist==0.1",
        "-t",
        target_dir,
        "--no-index",
        "--find-links",
        tmpdir,
    )
    result.did_create(Path("scratch") / "target" / "simpledist")


@pytest.mark.usefixtures("with_wheel")
def test_install_wheel_with_target_and_data_files(
    script: PipTestEnvironment, data: TestData
) -> None:
    """
    Test for issue #4092. It will be checked that a data_files specification in
    setup.py is handled correctly when a wheel is installed with the --target
    option.

    The setup() for the wheel 'prjwithdatafile-1.0-py2.py3-none-any.whl' is as
    follows ::

        setup(
            name='prjwithdatafile',
            version='1.0',
            packages=['prjwithdatafile'],
            data_files=[
                (r'packages1', ['prjwithdatafile/README.txt']),
                (r'packages2', ['prjwithdatafile/README.txt'])
            ]
        )
    """
    target_dir = script.scratch_path / "prjwithdatafile"
    package = data.packages.joinpath("prjwithdatafile-1.0-py2.py3-none-any.whl")
    result = script.pip("install", package, "-t", target_dir, "--no-index")
    project_path = Path("scratch") / "prjwithdatafile"
    result.did_create(project_path / "packages1" / "README.txt")
    result.did_create(project_path / "packages2" / "README.txt")
    result.did_not_create(project_path / "lib" / "python")


def test_install_wheel_with_root(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing a wheel using pip install --root
    """
    root_dir = script.scratch_path / "root"
    shutil.copy(shared_data.packages / "simple.dist-0.1-py2.py3-none-any.whl", tmpdir)
    result = script.pip(
        "install",
        "simple.dist==0.1",
        "--root",
        root_dir,
        "--no-index",
        "--find-links",
        tmpdir,
    )
    result.did_create(Path("scratch") / "root")


def test_install_wheel_with_prefix(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing a wheel using pip install --prefix
    """
    prefix_dir = script.scratch_path / "prefix"
    shutil.copy(shared_data.packages / "simple.dist-0.1-py2.py3-none-any.whl", tmpdir)
    result = script.pip(
        "install",
        "simple.dist==0.1",
        "--prefix",
        prefix_dir,
        "--no-index",
        "--find-links",
        tmpdir,
    )
    lib = distutils.sysconfig.get_python_lib(prefix=os.path.join("scratch", "prefix"))
    result.did_create(lib)


def test_install_from_wheel_installs_deps(
    script: PipTestEnvironment, data: TestData, tmpdir: Path
) -> None:
    """
    Test can install dependencies of wheels
    """
    # 'requires_source' depends on the 'source' project
    package = data.packages.joinpath("requires_source-1.0-py2.py3-none-any.whl")
    shutil.copy(data.packages / "source-1.0.tar.gz", tmpdir)
    result = script.pip(
        "install",
        "--no-index",
        "--find-links",
        tmpdir,
        package,
    )
    result.assert_installed("source", editable=False)


def test_install_from_wheel_no_deps(
    script: PipTestEnvironment, data: TestData, tmpdir: Path
) -> None:
    """
    Test --no-deps works with wheel installs
    """
    # 'requires_source' depends on the 'source' project
    package = data.packages.joinpath("requires_source-1.0-py2.py3-none-any.whl")
    shutil.copy(data.packages / "source-1.0.tar.gz", tmpdir)
    result = script.pip(
        "install",
        "--no-index",
        "--find-links",
        tmpdir,
        "--no-deps",
        package,
    )
    pkg_folder = script.site_packages / "source"
    result.did_not_create(pkg_folder)


def test_wheel_record_lines_in_deterministic_order(
    script: PipTestEnvironment, data: TestData
) -> None:
    to_install = data.packages.joinpath("simplewheel-1.0-py2.py3-none-any.whl")
    result = script.pip("install", to_install)

    dist_info_folder = script.site_packages / "simplewheel-1.0.dist-info"
    record_path = dist_info_folder / "RECORD"

    result.did_create(dist_info_folder)
    result.did_create(record_path)

    record_path = result.files_created[record_path].full
    record_lines = [p for p in Path(record_path).read_text().split("\n") if p]
    assert record_lines == sorted(record_lines)


def test_wheel_record_lines_have_hash_for_data_files(
    script: PipTestEnvironment,
) -> None:
    package = make_wheel(
        "simple",
        "0.1.0",
        extra_data_files={
            "purelib/info.txt": "c",
        },
    ).save_to_dir(script.scratch_path)
    script.pip("install", package)
    record_file = script.site_packages_path / "simple-0.1.0.dist-info" / "RECORD"
    record_text = record_file.read_text()
    record_rows = list(csv.reader(record_text.splitlines()))
    records = {r[0]: r[1:] for r in record_rows}
    assert records["info.txt"] == [
        "sha256=Ln0sA6lQeuJl7PW1NWiFpTOTogKdJBOUmXJloaJa78Y",
        "1",
    ]


def test_wheel_record_lines_have_updated_hash_for_scripts(
    script: PipTestEnvironment,
) -> None:
    """
    pip rewrites "#!python" shebang lines in scripts when it installs them;
    make sure it updates the RECORD file correspondingly.
    """
    package = make_wheel(
        "simple",
        "0.1.0",
        extra_data_files={
            "scripts/dostuff": "#!python\n",
        },
    ).save_to_dir(script.scratch_path)
    script.pip("install", package)
    record_file = script.site_packages_path / "simple-0.1.0.dist-info" / "RECORD"
    record_text = record_file.read_text()
    record_rows = list(csv.reader(record_text.splitlines()))
    records = {r[0]: r[1:] for r in record_rows}

    script_path = script.bin_path / "dostuff"
    script_contents = script_path.read_bytes()
    assert not script_contents.startswith(b"#!python\n")

    script_digest = hashlib.sha256(script_contents).digest()
    script_digest_b64 = (
        base64.urlsafe_b64encode(script_digest).decode("US-ASCII").rstrip("=")
    )

    script_record_path = os.path.relpath(
        script_path, script.site_packages_path
    ).replace(os.path.sep, "/")
    assert records[script_record_path] == [
        f"sha256={script_digest_b64}",
        str(len(script_contents)),
    ]


@pytest.mark.incompatible_with_test_venv
@pytest.mark.usefixtures("with_wheel")
def test_install_user_wheel(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test user install from wheel (that has a script)
    """
    shutil.copy(shared_data.packages / "has.script-1.0-py2.py3-none-any.whl", tmpdir)
    result = script.pip(
        "install",
        "has.script==1.0",
        "--user",
        "--no-index",
        "--find-links",
        tmpdir,
    )
    dist_info_folder = script.user_site / "has.script-1.0.dist-info"
    result.did_create(dist_info_folder)
    script_file = script.user_bin / "script.py"
    result.did_create(script_file)


def test_install_from_wheel_gen_entrypoint(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing scripts (entry points are generated)
    """
    shutil.copy(
        shared_data.packages / "script.wheel1a-0.1-py2.py3-none-any.whl",
        tmpdir,
    )
    result = script.pip(
        "install",
        "script.wheel1a==0.1",
        "--no-index",
        "--find-links",
        tmpdir,
    )
    if os.name == "nt":
        wrapper_file = script.bin / "t1.exe"
    else:
        wrapper_file = script.bin / "t1"
    result.did_create(wrapper_file)

    if os.name != "nt":
        assert bool(os.access(script.base_path / wrapper_file, os.X_OK))


def test_install_from_wheel_gen_uppercase_entrypoint(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing scripts with uppercase letters in entry point names
    """
    shutil.copy(
        shared_data.packages / "console_scripts_uppercase-1.0-py2.py3-none-any.whl",
        tmpdir,
    )
    result = script.pip(
        "install",
        "console-scripts-uppercase==1.0",
        "--no-index",
        "--find-links",
        tmpdir,
    )
    if os.name == "nt":
        # Case probably doesn't make any difference on NT
        wrapper_file = script.bin / "cmdName.exe"
    else:
        wrapper_file = script.bin / "cmdName"
    result.did_create(wrapper_file)

    if os.name != "nt":
        assert bool(os.access(script.base_path / wrapper_file, os.X_OK))


def test_install_from_wheel_gen_unicode_entrypoint(script: PipTestEnvironment) -> None:
    make_wheel(
        "script_wheel_unicode",
        "1.0",
        console_scripts=["進入點 = 模組:函式"],
    ).save_to_dir(script.scratch_path)

    result = script.pip(
        "install",
        "--no-index",
        "--find-links",
        script.scratch_path,
        "script_wheel_unicode",
    )
    if os.name == "nt":
        result.did_create(script.bin.joinpath("進入點.exe"))
    else:
        result.did_create(script.bin.joinpath("進入點"))


def test_install_from_wheel_with_legacy(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing scripts (legacy scripts are preserved)
    """
    shutil.copy(
        shared_data.packages / "script.wheel2a-0.1-py2.py3-none-any.whl",
        tmpdir,
    )
    result = script.pip(
        "install",
        "script.wheel2a==0.1",
        "--no-index",
        "--find-links",
        tmpdir,
    )

    legacy_file1 = script.bin / "testscript1.bat"
    legacy_file2 = script.bin / "testscript2"

    result.did_create(legacy_file1)
    result.did_create(legacy_file2)


def test_install_from_wheel_no_setuptools_entrypoint(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test that when we generate scripts, any existing setuptools wrappers in
    the wheel are skipped.
    """
    shutil.copy(shared_data.packages / "script.wheel1-0.1-py2.py3-none-any.whl", tmpdir)
    result = script.pip(
        "install",
        "script.wheel1==0.1",
        "--no-index",
        "--find-links",
        tmpdir,
    )
    if os.name == "nt":
        wrapper_file = script.bin / "t1.exe"
    else:
        wrapper_file = script.bin / "t1"
    wrapper_helper = script.bin / "t1-script.py"

    # The wheel has t1.exe and t1-script.py. We will be generating t1 or
    # t1.exe depending on the platform. So we check that the correct wrapper
    # is present and that the -script.py helper has been skipped. We can't
    # easily test that the wrapper from the wheel has been skipped /
    # overwritten without getting very platform-dependent, so omit that.
    result.did_create(wrapper_file)
    result.did_not_create(wrapper_helper)


def test_skipping_setuptools_doesnt_skip_legacy(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing scripts (legacy scripts are preserved even when we skip
    setuptools wrappers)
    """
    shutil.copy(shared_data.packages / "script.wheel2-0.1-py2.py3-none-any.whl", tmpdir)
    result = script.pip(
        "install",
        "script.wheel2==0.1",
        "--no-index",
        "--find-links",
        tmpdir,
    )

    legacy_file1 = script.bin / "testscript1.bat"
    legacy_file2 = script.bin / "testscript2"
    wrapper_helper = script.bin / "t1-script.py"

    result.did_create(legacy_file1)
    result.did_create(legacy_file2)
    result.did_not_create(wrapper_helper)


def test_install_from_wheel_gui_entrypoint(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing scripts (gui entry points are generated)
    """
    shutil.copy(shared_data.packages / "script.wheel3-0.1-py2.py3-none-any.whl", tmpdir)
    result = script.pip(
        "install",
        "script.wheel3==0.1",
        "--no-index",
        "--find-links",
        tmpdir,
    )
    if os.name == "nt":
        wrapper_file = script.bin / "t1.exe"
    else:
        wrapper_file = script.bin / "t1"
    result.did_create(wrapper_file)


def test_wheel_compiles_pyc(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing from wheel with --compile on
    """
    shutil.copy(shared_data.packages / "simple.dist-0.1-py2.py3-none-any.whl", tmpdir)
    script.pip(
        "install",
        "--compile",
        "simple.dist==0.1",
        "--no-index",
        "--find-links",
        tmpdir,
    )
    # There are many locations for the __init__.pyc file so attempt to find
    #   any of them
    exists = [
        os.path.exists(script.site_packages_path / "simpledist/__init__.pyc"),
        *script.site_packages_path.glob("simpledist/__pycache__/__init__*.pyc"),
    ]
    assert any(exists)


def test_wheel_no_compiles_pyc(
    script: PipTestEnvironment, shared_data: TestData, tmpdir: Path
) -> None:
    """
    Test installing from wheel with --compile on
    """
    shutil.copy(shared_data.packages / "simple.dist-0.1-py2.py3-none-any.whl", tmpdir)
    script.pip(
        "install",
        "--no-compile",
        "simple.dist==0.1",
        "--no-index",
        "--find-links",
        tmpdir,
    )
    # There are many locations for the __init__.pyc file so attempt to find
    #   any of them
    exists = [
        os.path.exists(script.site_packages_path / "simpledist/__init__.pyc"),
        *script.site_packages_path.glob("simpledist/__pycache__/__init__*.pyc"),
    ]

    assert not any(exists)


def test_install_from_wheel_uninstalls_old_version(
    script: PipTestEnvironment, data: TestData
) -> None:
    # regression test for https://github.com/pypa/pip/issues/1825
    package = data.packages.joinpath("simplewheel-1.0-py2.py3-none-any.whl")
    result = script.pip("install", package, "--no-index")
    package = data.packages.joinpath("simplewheel-2.0-py2.py3-none-any.whl")
    result = script.pip("install", package, "--no-index")
    dist_info_folder = script.site_packages / "simplewheel-2.0.dist-info"
    result.did_create(dist_info_folder)
    dist_info_folder = script.site_packages / "simplewheel-1.0.dist-info"
    result.did_not_create(dist_info_folder)


def test_wheel_compile_syntax_error(script: PipTestEnvironment, data: TestData) -> None:
    package = data.packages.joinpath("compilewheel-1.0-py2.py3-none-any.whl")
    result = script.pip("install", "--compile", package, "--no-index")
    assert "yield from" not in result.stdout
    assert "SyntaxError: " not in result.stdout


def test_wheel_install_with_no_cache_dir(
    script: PipTestEnvironment, data: TestData
) -> None:
    """Check wheel installations work, even with no cache."""
    package = data.packages.joinpath("simple.dist-0.1-py2.py3-none-any.whl")
    result = script.pip("install", "--no-cache-dir", "--no-index", package)
    result.assert_installed("simpledist", editable=False)


def test_wheel_install_fails_with_extra_dist_info(script: PipTestEnvironment) -> None:
    package = create_basic_wheel_for_package(
        script,
        "simple",
        "0.1.0",
        extra_files={
            "unrelated-2.0.0.dist-info/WHEEL": "Wheel-Version: 1.0",
            "unrelated-2.0.0.dist-info/METADATA": ("Name: unrelated\nVersion: 2.0.0\n"),
        },
    )
    result = script.pip(
        "install", "--no-cache-dir", "--no-index", package, expect_error=True
    )
    assert "multiple .dist-info directories" in result.stderr


def test_wheel_install_fails_with_unrelated_dist_info(
    script: PipTestEnvironment,
) -> None:
    package = create_basic_wheel_for_package(script, "simple", "0.1.0")
    new_name = "unrelated-2.0.0-py2.py3-none-any.whl"
    new_package = os.path.join(os.path.dirname(package), new_name)
    shutil.move(os.fspath(package), new_package)

    result = script.pip(
        "install",
        "--no-cache-dir",
        "--no-index",
        new_package,
        expect_error=True,
    )

    assert "'simple-0.1.0.dist-info' does not start with 'unrelated'" in result.stderr


def test_wheel_installs_ok_with_nested_dist_info(script: PipTestEnvironment) -> None:
    package = create_basic_wheel_for_package(
        script,
        "simple",
        "0.1.0",
        extra_files={
            "subdir/unrelated-2.0.0.dist-info/WHEEL": "Wheel-Version: 1.0",
            "subdir/unrelated-2.0.0.dist-info/METADATA": (
                "Name: unrelated\nVersion: 2.0.0\n"
            ),
        },
    )
    script.pip("install", "--no-cache-dir", "--no-index", package)


def test_wheel_installs_ok_with_badly_encoded_irrelevant_dist_info_file(
    script: PipTestEnvironment,
) -> None:
    package = create_basic_wheel_for_package(
        script,
        "simple",
        "0.1.0",
        extra_files={"simple-0.1.0.dist-info/AUTHORS.txt": b"\xff"},
    )
    script.pip("install", "--no-cache-dir", "--no-index", package)


def test_wheel_install_fails_with_badly_encoded_metadata(
    script: PipTestEnvironment,
) -> None:
    package = create_basic_wheel_for_package(
        script,
        "simple",
        "0.1.0",
        extra_files={"simple-0.1.0.dist-info/METADATA": b"\xff"},
    )
    result = script.pip(
        "install", "--no-cache-dir", "--no-index", package, expect_error=True
    )
    assert "Error decoding metadata for" in result.stderr
    assert "simple-0.1.0-py2.py3-none-any.whl" in result.stderr
    assert "METADATA" in result.stderr


@pytest.mark.parametrize(
    "package_name",
    ["simple-package", "simple_package"],
)
def test_correct_package_name_while_creating_wheel_bug(
    script: PipTestEnvironment, package_name: str
) -> None:
    """Check that the package name is correctly named while creating
    a .whl file with a given format
    """
    package = create_basic_wheel_for_package(script, package_name, "1.0")
    wheel_name = os.path.basename(package)
    assert wheel_name == "simple_package-1.0-py2.py3-none-any.whl"


@pytest.mark.parametrize("name", ["purelib", "abc"])
def test_wheel_with_file_in_data_dir_has_reasonable_error(
    script: PipTestEnvironment, tmpdir: Path, name: str
) -> None:
    """Normally we expect entities in the .data directory to be in a
    subdirectory, but if they are not then we should show a reasonable error
    message that includes the path.
    """
    wheel_path = make_wheel(
        "simple", "0.1.0", extra_data_files={name: "hello world"}
    ).save_to_dir(tmpdir)

    result = script.pip("install", "--no-index", str(wheel_path), expect_error=True)
    assert f"simple-0.1.0.data/{name}" in result.stderr


def test_wheel_with_unknown_subdir_in_data_dir_has_reasonable_error(
    script: PipTestEnvironment, tmpdir: Path
) -> None:
    wheel_path = make_wheel(
        "simple", "0.1.0", extra_data_files={"unknown/hello.txt": "hello world"}
    ).save_to_dir(tmpdir)

    result = script.pip("install", "--no-index", str(wheel_path), expect_error=True)
    assert "simple-0.1.0.data/unknown/hello.txt" in result.stderr



__version__: str

from .providers import AbstractProvider as AbstractProvider
from .providers import AbstractResolver as AbstractResolver
from .reporters import BaseReporter as BaseReporter
from .resolvers import InconsistentCandidate as InconsistentCandidate
from .resolvers import RequirementsConflicted as RequirementsConflicted
from .resolvers import ResolutionError as ResolutionError
from .resolvers import ResolutionImpossible as ResolutionImpossible
from .resolvers import ResolutionTooDeep as ResolutionTooDeep
from .resolvers import Resolver as Resolver



import os
from contextlib import contextmanager
from optparse import Values
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Iterator, List, Tuple, Union, cast

import pytest

import pip._internal.configuration
from pip._internal.cli.main import main
from pip._internal.commands import create_command
from pip._internal.commands.configuration import ConfigurationCommand
from pip._internal.exceptions import PipError
from tests.lib.options_helpers import AddFakeCommandMixin


@contextmanager
def assert_option_error(
    capsys: pytest.CaptureFixture[str], expected: str
) -> Iterator[None]:
    """
    Assert that a SystemExit occurred because of a parsing error.

    Args:
      expected: an expected substring of stderr.
    """
    with pytest.raises(SystemExit) as excinfo:
        yield

    assert excinfo.value.code == 2
    stderr = capsys.readouterr().err
    assert expected in stderr


def assert_is_default_cache_dir(value: str) -> None:
    # This path looks different on different platforms, but the path always
    # has the substring "pip".
    assert "pip" in value


class TestOptionPrecedence(AddFakeCommandMixin):
    """
    Tests for confirming our option precedence:
        cli -> environment -> subcommand config -> global config -> option
        defaults
    """

    def get_config_section(self, section: str) -> List[Tuple[str, str]]:
        config = {
            "global": [("timeout", "-3")],
            "fake": [("timeout", "-2")],
        }
        return config[section]

    def get_config_section_global(self, section: str) -> List[Tuple[str, str]]:
        config: Dict[str, List[Tuple[str, str]]] = {
            "global": [("timeout", "-3")],
            "fake": [],
        }
        return config[section]

    def test_env_override_default_int(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """
        Test that environment variable overrides an int option default.
        """
        monkeypatch.setenv("PIP_TIMEOUT", "-1")
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake"]))
        assert options.timeout == -1

    @pytest.mark.parametrize("values", (["F1"], ["F1", "F2"]))
    def test_env_override_default_append(
        self, values: List[str], monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        Test that environment variable overrides an append option default.
        """
        monkeypatch.setenv("PIP_FIND_LINKS", " ".join(values))
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake"]))
        assert options.find_links == values

    @pytest.mark.parametrize("choices", (["w"], ["s", "w"]))
    def test_env_override_default_choice(
        self, choices: List[str], monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        Test that environment variable overrides a choice option default.
        """
        monkeypatch.setenv("PIP_EXISTS_ACTION", " ".join(choices))
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake"]))
        assert options.exists_action == choices

    @pytest.mark.parametrize("name", ("PIP_LOG_FILE", "PIP_LOCAL_LOG"))
    def test_env_alias_override_default(
        self, name: str, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        When an option has multiple long forms, test that the technique of
        using the env variable, "PIP_<long form>" works for all cases.
        (e.g. PIP_LOG_FILE and PIP_LOCAL_LOG should all work)
        """
        monkeypatch.setenv(name, "override.log")
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake"]))
        assert options.log == "override.log"

    def test_cli_override_environment(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """
        Test the cli overrides and environment variable
        """
        monkeypatch.setenv("PIP_TIMEOUT", "-1")
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["fake", "--timeout", "-2"])
        )
        assert options.timeout == -2

    @pytest.mark.parametrize(
        "pip_no_cache_dir",
        [
            # Enabling --no-cache-dir means no cache directory.
            "1",
            "true",
            "on",
            "yes",
            # For historical / backwards compatibility reasons, we also disable
            # the cache directory if provided a value that translates to 0.
            "0",
            "false",
            "off",
            "no",
        ],
    )
    def test_cache_dir__PIP_NO_CACHE_DIR(
        self, pip_no_cache_dir: str, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        Test setting the PIP_NO_CACHE_DIR environment variable without
        passing any command-line flags.
        """
        monkeypatch.setenv("PIP_NO_CACHE_DIR", pip_no_cache_dir)
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake"]))
        assert options.cache_dir is False

    @pytest.mark.parametrize("pip_no_cache_dir", ["yes", "no"])
    def test_cache_dir__PIP_NO_CACHE_DIR__with_cache_dir(
        self,
        pip_no_cache_dir: str,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """
        Test setting PIP_NO_CACHE_DIR while also passing an explicit
        --cache-dir value.
        """
        monkeypatch.setenv("PIP_NO_CACHE_DIR", pip_no_cache_dir)
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["--cache-dir", "/cache/dir", "fake"])
        )
        # The command-line flag takes precedence.
        assert options.cache_dir == "/cache/dir"

    @pytest.mark.parametrize("pip_no_cache_dir", ["yes", "no"])
    def test_cache_dir__PIP_NO_CACHE_DIR__with_no_cache_dir(
        self,
        pip_no_cache_dir: str,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """
        Test setting PIP_NO_CACHE_DIR while also passing --no-cache-dir.
        """
        monkeypatch.setenv("PIP_NO_CACHE_DIR", pip_no_cache_dir)
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["--no-cache-dir", "fake"]))
        # The command-line flag should take precedence (which has the same
        # value in this case).
        assert options.cache_dir is False

    def test_cache_dir__PIP_NO_CACHE_DIR_invalid__with_no_cache_dir(
        self,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """
        Test setting PIP_NO_CACHE_DIR to an invalid value while also passing
        --no-cache-dir.
        """
        monkeypatch.setenv("PIP_NO_CACHE_DIR", "maybe")
        expected_err = "--no-cache-dir error: invalid truth value 'maybe'"
        with assert_option_error(capsys, expected=expected_err):
            main(["--no-cache-dir", "fake"])


class TestUsePEP517Options:

    """
    Test options related to using --use-pep517.
    """

    def parse_args(self, args: List[str]) -> Values:
        # We use DownloadCommand since that is one of the few Command
        # classes with the use_pep517 options.
        command = create_command("download")
        options, args = command.parse_args(args)

        return options

    def test_no_option(self) -> None:
        """
        Test passing no option.
        """
        options = self.parse_args([])
        assert options.use_pep517 is None

    def test_use_pep517(self) -> None:
        """
        Test passing --use-pep517.
        """
        options = self.parse_args(["--use-pep517"])
        assert options.use_pep517 is True

    def test_no_use_pep517(self) -> None:
        """
        Test passing --no-use-pep517.
        """
        options = self.parse_args(["--no-use-pep517"])
        assert options.use_pep517 is False

    def test_PIP_USE_PEP517_true(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """
        Test setting PIP_USE_PEP517 to "true".
        """
        monkeypatch.setenv("PIP_USE_PEP517", "true")
        options = self.parse_args([])
        # This is an int rather than a boolean because strtobool() in pip's
        # configuration code returns an int.
        assert options.use_pep517 == 1

    def test_PIP_USE_PEP517_false(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """
        Test setting PIP_USE_PEP517 to "false".
        """
        monkeypatch.setenv("PIP_USE_PEP517", "false")
        options = self.parse_args([])
        # This is an int rather than a boolean because strtobool() in pip's
        # configuration code returns an int.
        assert options.use_pep517 == 0

    def test_use_pep517_and_PIP_USE_PEP517_false(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        Test passing --use-pep517 and setting PIP_USE_PEP517 to "false".
        """
        monkeypatch.setenv("PIP_USE_PEP517", "false")
        options = self.parse_args(["--use-pep517"])
        assert options.use_pep517 is True

    def test_no_use_pep517_and_PIP_USE_PEP517_true(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        Test passing --no-use-pep517 and setting PIP_USE_PEP517 to "true".
        """
        monkeypatch.setenv("PIP_USE_PEP517", "true")
        options = self.parse_args(["--no-use-pep517"])
        assert options.use_pep517 is False

    def test_PIP_NO_USE_PEP517(
        self, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
    ) -> None:
        """
        Test setting PIP_NO_USE_PEP517, which isn't allowed.
        """
        monkeypatch.setenv("PIP_NO_USE_PEP517", "true")
        with assert_option_error(capsys, expected="--no-use-pep517 error"):
            self.parse_args([])


class TestOptionsInterspersed(AddFakeCommandMixin):
    def test_general_option_after_subcommand(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["fake", "--timeout", "-1"])
        )
        assert options.timeout == -1

    def test_option_after_subcommand_arg(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["fake", "arg", "--timeout", "-1"])
        )
        assert options.timeout == -1

    def test_additive_before_after_subcommand(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["-v", "fake", "-v"]))
        assert options.verbose == 2

    def test_subcommand_option_before_subcommand_fails(self) -> None:
        with pytest.raises(SystemExit):
            main(["--find-links", "F1", "fake"])


@contextmanager
def tmpconfig(option: str, value: Any, section: str = "global") -> Iterator[str]:
    with NamedTemporaryFile(mode="w", delete=False) as f:
        f.write(f"[{section}]\n{option}={value}\n")
        name = f.name
    try:
        yield name
    finally:
        os.unlink(name)


class TestCountOptions(AddFakeCommandMixin):
    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", range(4))
    def test_cli_long(self, option: str, value: int) -> None:
        flags = [f"--{option}"] * value
        # FakeCommand intentionally returns the wrong type.
        opt1, args1 = cast(Tuple[Values, List[str]], main(flags + ["fake"]))
        opt2, args2 = cast(Tuple[Values, List[str]], main(["fake"] + flags))
        assert getattr(opt1, option) == getattr(opt2, option) == value

    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", range(1, 4))
    def test_cli_short(self, option: str, value: int) -> None:
        flag = "-" + option[0] * value
        # FakeCommand intentionally returns the wrong type.
        opt1, args1 = cast(Tuple[Values, List[str]], main([flag, "fake"]))
        opt2, args2 = cast(Tuple[Values, List[str]], main(["fake", flag]))
        assert getattr(opt1, option) == getattr(opt2, option) == value

    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", range(4))
    def test_env_var(
        self, option: str, value: int, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setenv("PIP_" + option.upper(), str(value))
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake"]))
        assert getattr(options, option) == value

    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", range(3))
    def test_env_var_integrate_cli(
        self, option: str, value: int, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setenv("PIP_" + option.upper(), str(value))
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake", "--" + option]))
        assert getattr(options, option) == value + 1

    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", (-1, "foobar"))
    def test_env_var_invalid(
        self,
        option: str,
        value: Any,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        monkeypatch.setenv("PIP_" + option.upper(), str(value))
        with assert_option_error(capsys, expected="a non-negative integer"):
            main(["fake"])

    # Undocumented, support for backward compatibility
    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", ("no", "false"))
    def test_env_var_false(
        self, option: str, value: str, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setenv("PIP_" + option.upper(), str(value))
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake"]))
        assert getattr(options, option) == 0

    # Undocumented, support for backward compatibility
    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", ("yes", "true"))
    def test_env_var_true(
        self, option: str, value: str, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setenv("PIP_" + option.upper(), str(value))
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake"]))
        assert getattr(options, option) == 1

    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", range(4))
    def test_config_file(
        self, option: str, value: int, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        with tmpconfig(option, value) as name:
            monkeypatch.setenv("PIP_CONFIG_FILE", name)
            # FakeCommand intentionally returns the wrong type.
            options, args = cast(Tuple[Values, List[str]], main(["fake"]))
            assert getattr(options, option) == value

    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", range(3))
    def test_config_file_integrate_cli(
        self, option: str, value: int, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        with tmpconfig(option, value) as name:
            monkeypatch.setenv("PIP_CONFIG_FILE", name)
            # FakeCommand intentionally returns the wrong type.
            options, args = cast(
                Tuple[Values, List[str]], main(["fake", "--" + option])
            )
            assert getattr(options, option) == value + 1

    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", (-1, "foobar"))
    def test_config_file_invalid(
        self,
        option: str,
        value: Any,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        with tmpconfig(option, value) as name:
            monkeypatch.setenv("PIP_CONFIG_FILE", name)
            with assert_option_error(capsys, expected="non-negative integer"):
                main(["fake"])

    # Undocumented, support for backward compatibility
    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", ("no", "false"))
    def test_config_file_false(
        self, option: str, value: str, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        with tmpconfig(option, value) as name:
            monkeypatch.setenv("PIP_CONFIG_FILE", name)
            # FakeCommand intentionally returns the wrong type.
            options, args = cast(Tuple[Values, List[str]], main(["fake"]))
            assert getattr(options, option) == 0

    # Undocumented, support for backward compatibility
    @pytest.mark.parametrize("option", ("verbose", "quiet"))
    @pytest.mark.parametrize("value", ("yes", "true"))
    def test_config_file_true(
        self, option: str, value: str, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        with tmpconfig(option, value) as name:
            monkeypatch.setenv("PIP_CONFIG_FILE", name)
            # FakeCommand intentionally returns the wrong type.
            options, args = cast(Tuple[Values, List[str]], main(["fake"]))
            assert getattr(options, option) == 1


class TestGeneralOptions(AddFakeCommandMixin):

    # the reason to specifically test general options is due to the
    # extra processing they receive, and the number of bugs we've had

    def test_cache_dir__default(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["fake"]))
        # With no options the default cache dir should be used.
        assert_is_default_cache_dir(options.cache_dir)

    def test_cache_dir__provided(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["--cache-dir", "/cache/dir", "fake"])
        )
        assert options.cache_dir == "/cache/dir"

    def test_no_cache_dir__provided(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(Tuple[Values, List[str]], main(["--no-cache-dir", "fake"]))
        assert options.cache_dir is False

    def test_require_virtualenv(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(
            Tuple[Values, List[str]], main(["--require-virtualenv", "fake"])
        )
        options2, args2 = cast(
            Tuple[Values, List[str]], main(["fake", "--require-virtualenv"])
        )
        assert options1.require_venv
        assert options2.require_venv

    def test_log(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(
            Tuple[Values, List[str]], main(["--log", "path", "fake"])
        )
        options2, args2 = cast(
            Tuple[Values, List[str]], main(["fake", "--log", "path"])
        )
        assert options1.log == options2.log == "path"

    def test_local_log(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(
            Tuple[Values, List[str]], main(["--local-log", "path", "fake"])
        )
        options2, args2 = cast(
            Tuple[Values, List[str]], main(["fake", "--local-log", "path"])
        )
        assert options1.log == options2.log == "path"

    def test_no_input(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(Tuple[Values, List[str]], main(["--no-input", "fake"]))
        options2, args2 = cast(Tuple[Values, List[str]], main(["fake", "--no-input"]))
        assert options1.no_input
        assert options2.no_input

    def test_proxy(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(
            Tuple[Values, List[str]], main(["--proxy", "path", "fake"])
        )
        options2, args2 = cast(
            Tuple[Values, List[str]], main(["fake", "--proxy", "path"])
        )
        assert options1.proxy == options2.proxy == "path"

    def test_retries(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(
            Tuple[Values, List[str]], main(["--retries", "-1", "fake"])
        )
        options2, args2 = cast(
            Tuple[Values, List[str]], main(["fake", "--retries", "-1"])
        )
        assert options1.retries == options2.retries == -1

    def test_timeout(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(
            Tuple[Values, List[str]], main(["--timeout", "-1", "fake"])
        )
        options2, args2 = cast(
            Tuple[Values, List[str]], main(["fake", "--timeout", "-1"])
        )
        assert options1.timeout == options2.timeout == -1

    def test_exists_action(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(
            Tuple[Values, List[str]], main(["--exists-action", "w", "fake"])
        )
        options2, args2 = cast(
            Tuple[Values, List[str]], main(["fake", "--exists-action", "w"])
        )
        assert options1.exists_action == options2.exists_action == ["w"]

    def test_cert(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(
            Tuple[Values, List[str]], main(["--cert", "path", "fake"])
        )
        options2, args2 = cast(
            Tuple[Values, List[str]], main(["fake", "--cert", "path"])
        )
        assert options1.cert == options2.cert == "path"

    def test_client_cert(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options1, args1 = cast(
            Tuple[Values, List[str]], main(["--client-cert", "path", "fake"])
        )
        options2, args2 = cast(
            Tuple[Values, List[str]], main(["fake", "--client-cert", "path"])
        )
        assert options1.client_cert == options2.client_cert == "path"


class TestOptionsConfigFiles:
    def test_venv_config_file_found(self, monkeypatch: pytest.MonkeyPatch) -> None:
        # strict limit on the global config files list
        monkeypatch.setattr(
            pip._internal.utils.appdirs, "site_config_dirs", lambda _: ["/a/place"]
        )

        cp = pip._internal.configuration.Configuration(isolated=False)

        files = []
        for _, val in cp.iter_config_files():
            files.extend(val)

        assert len(files) == 4

    @pytest.mark.parametrize(
        "args, expect",
        (
            ([], None),
            (["--global"], "global"),
            (["--site"], "site"),
            (["--user"], "user"),
            (["--global", "--user"], PipError),
            (["--global", "--site"], PipError),
            (["--global", "--site", "--user"], PipError),
        ),
    )
    def test_config_file_options(
        self,
        monkeypatch: pytest.MonkeyPatch,
        args: List[str],
        expect: Union[None, str, PipError],
    ) -> None:
        cmd = cast(ConfigurationCommand, create_command("config"))
        # Replace a handler with a no-op to avoid side effects
        monkeypatch.setattr(cmd, "get_name", lambda *a: None)

        options, args = cmd.parser.parse_args(args + ["get", "name"])
        if expect is PipError:
            with pytest.raises(PipError):
                cmd._determine_file(options, need_value=False)
        else:
            assert expect == cmd._determine_file(options, need_value=False)


class TestOptionsExpandUser(AddFakeCommandMixin):
    def test_cache_dir(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["--cache-dir", "~/cache/dir", "fake"])
        )
        assert options.cache_dir == os.path.expanduser("~/cache/dir")

    def test_log(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["--log", "~/path", "fake"])
        )
        assert options.log == os.path.expanduser("~/path")

    def test_local_log(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["--local-log", "~/path", "fake"])
        )
        assert options.log == os.path.expanduser("~/path")

    def test_cert(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["--cert", "~/path", "fake"])
        )
        assert options.cert == os.path.expanduser("~/path")

    def test_client_cert(self) -> None:
        # FakeCommand intentionally returns the wrong type.
        options, args = cast(
            Tuple[Values, List[str]], main(["--client-cert", "~/path", "fake"])
        )
        assert options.client_cert == os.path.expanduser("~/path")



"""
Contains functional tests of the Git class.
"""
import logging
import os
import pathlib
from typing import List, Optional, Tuple
from unittest.mock import Mock, patch

import pytest

from pip._internal.utils.misc import HiddenText
from pip._internal.vcs import vcs
from pip._internal.vcs.git import Git, RemoteNotFoundError
from tests.lib import PipTestEnvironment, _create_test_package, _git_commit


def test_get_backend_for_scheme() -> None:
    assert vcs.get_backend_for_scheme("git+https") is vcs.get_backend("Git")


def get_head_sha(script: PipTestEnvironment, dest: str) -> str:
    """Return the HEAD sha."""
    result = script.run("git", "rev-parse", "HEAD", cwd=dest)
    sha = result.stdout.strip()

    return sha


def checkout_ref(script: PipTestEnvironment, repo_dir: str, ref: str) -> None:
    script.run("git", "checkout", ref, cwd=repo_dir)


def checkout_new_branch(script: PipTestEnvironment, repo_dir: str, branch: str) -> None:
    script.run(
        "git",
        "checkout",
        "-b",
        branch,
        cwd=repo_dir,
    )


def do_commit(script: PipTestEnvironment, dest: str) -> str:
    _git_commit(script, dest, message="test commit", allow_empty=True)
    return get_head_sha(script, dest)


def add_commits(script: PipTestEnvironment, dest: str, count: int) -> List[str]:
    """Return a list of the commit hashes from oldest to newest."""
    shas = []
    for _ in range(count):
        sha = do_commit(script, dest)
        shas.append(sha)

    return shas


def check_rev(repo_dir: str, rev: str, expected: Tuple[Optional[str], bool]) -> None:
    assert Git.get_revision_sha(repo_dir, rev) == expected


def test_git_dir_ignored(tmpdir: pathlib.Path) -> None:
    """
    Test that a GIT_DIR environment variable is ignored.
    """
    repo_path = tmpdir / "test-repo"
    repo_path.mkdir()
    repo_dir = str(repo_path)

    env = {"GIT_DIR": "foo"}
    # If GIT_DIR is not ignored, then os.listdir() will return ['foo'].
    Git.run_command(["init", repo_dir], cwd=repo_dir, extra_environ=env)
    assert os.listdir(repo_dir) == [".git"]


def test_git_work_tree_ignored(tmpdir: pathlib.Path) -> None:
    """
    Test that a GIT_WORK_TREE environment variable is ignored.
    """
    repo_path = tmpdir / "test-repo"
    repo_path.mkdir()
    repo_dir = str(repo_path)

    Git.run_command(["init", repo_dir], cwd=repo_dir)
    # Choose a directory relative to the cwd that does not exist.
    # If GIT_WORK_TREE is not ignored, then the command will error out
    # with: "fatal: This operation must be run in a work tree".
    env = {"GIT_WORK_TREE": "foo"}
    Git.run_command(["status", repo_dir], extra_environ=env, cwd=repo_dir)


def test_get_remote_url(script: PipTestEnvironment, tmpdir: pathlib.Path) -> None:
    source_path = tmpdir / "source"
    source_path.mkdir()
    source_url = source_path.as_uri()

    source_dir = str(source_path)
    script.run("git", "init", cwd=source_dir)
    do_commit(script, source_dir)

    repo_dir = str(tmpdir / "repo")
    script.run("git", "clone", source_url, repo_dir)

    remote_url = Git.get_remote_url(repo_dir)
    assert remote_url == source_url


def test_get_remote_url__no_remote(
    script: PipTestEnvironment, tmpdir: pathlib.Path
) -> None:
    """
    Test a repo with no remote.
    """
    repo_path = tmpdir / "temp-repo"
    repo_path.mkdir()
    repo_dir = str(repo_path)

    script.run("git", "init", cwd=repo_dir)

    with pytest.raises(RemoteNotFoundError):
        Git.get_remote_url(repo_dir)


def test_get_current_branch(script: PipTestEnvironment) -> None:
    repo_dir = str(script.scratch_path)

    script.run("git", "init", cwd=repo_dir)
    sha = do_commit(script, repo_dir)

    assert Git.get_current_branch(repo_dir) == "master"

    # Switch to a branch with the same SHA as "master" but whose name
    # is alphabetically after.
    checkout_new_branch(script, repo_dir, "release")
    assert Git.get_current_branch(repo_dir) == "release"

    # Also test the detached HEAD case.
    checkout_ref(script, repo_dir, sha)
    assert Git.get_current_branch(repo_dir) is None


def test_get_current_branch__branch_and_tag_same_name(
    script: PipTestEnvironment, tmpdir: pathlib.Path
) -> None:
    """
    Check calling get_current_branch() from a branch or tag when the branch
    and tag have the same name.
    """
    repo_dir = str(tmpdir)
    script.run("git", "init", cwd=repo_dir)
    do_commit(script, repo_dir)
    checkout_new_branch(script, repo_dir, "dev")
    # Create a tag with the same name as the branch.
    script.run("git", "tag", "dev", cwd=repo_dir)

    assert Git.get_current_branch(repo_dir) == "dev"

    # Now try with the tag checked out.
    checkout_ref(script, repo_dir, "refs/tags/dev")
    assert Git.get_current_branch(repo_dir) is None


def test_get_revision_sha(script: PipTestEnvironment) -> None:
    repo_dir = str(script.scratch_path)

    script.run("git", "init", cwd=repo_dir)
    shas = add_commits(script, repo_dir, count=3)

    tag_sha = shas[0]
    origin_sha = shas[1]
    head_sha = shas[2]
    assert head_sha == shas[-1]

    origin_ref = "refs/remotes/origin/origin-branch"
    generic_ref = "refs/generic-ref"

    script.run("git", "branch", "local-branch", head_sha, cwd=repo_dir)
    script.run("git", "tag", "v1.0", tag_sha, cwd=repo_dir)
    script.run("git", "update-ref", origin_ref, origin_sha, cwd=repo_dir)
    script.run(
        "git",
        "update-ref",
        "refs/remotes/upstream/upstream-branch",
        head_sha,
        cwd=repo_dir,
    )
    script.run("git", "update-ref", generic_ref, head_sha, cwd=repo_dir)

    # Test two tags pointing to the same sha.
    script.run("git", "tag", "v2.0", tag_sha, cwd=repo_dir)
    # Test tags sharing the same suffix as another tag, both before and
    # after the suffix alphabetically.
    script.run("git", "tag", "aaa/v1.0", head_sha, cwd=repo_dir)
    script.run("git", "tag", "zzz/v1.0", head_sha, cwd=repo_dir)

    check_rev(repo_dir, "v1.0", (tag_sha, False))
    check_rev(repo_dir, "v2.0", (tag_sha, False))
    check_rev(repo_dir, "origin-branch", (origin_sha, True))

    ignored_names = [
        # Local branches should be ignored.
        "local-branch",
        # Non-origin remote branches should be ignored.
        "upstream-branch",
        # Generic refs should be ignored.
        "generic-ref",
        # Fully spelled-out refs should be ignored.
        origin_ref,
        generic_ref,
        # Test passing a valid commit hash.
        tag_sha,
        # Test passing a non-existent name.
        "does-not-exist",
    ]
    for name in ignored_names:
        check_rev(repo_dir, name, (None, False))


def test_is_commit_id_equal(script: PipTestEnvironment) -> None:
    """
    Test Git.is_commit_id_equal().
    """
    version_pkg_path = os.fspath(_create_test_package(script.scratch_path))
    script.run("git", "branch", "branch0.1", cwd=version_pkg_path)
    commit = script.run("git", "rev-parse", "HEAD", cwd=version_pkg_path).stdout.strip()

    assert Git.is_commit_id_equal(version_pkg_path, commit)
    assert not Git.is_commit_id_equal(version_pkg_path, commit[:7])
    assert not Git.is_commit_id_equal(version_pkg_path, "branch0.1")
    assert not Git.is_commit_id_equal(version_pkg_path, "abc123")
    # Also check passing a None value.
    assert not Git.is_commit_id_equal(version_pkg_path, None)


def test_is_immutable_rev_checkout(script: PipTestEnvironment) -> None:
    version_pkg_path = os.fspath(_create_test_package(script.scratch_path))
    commit = script.run("git", "rev-parse", "HEAD", cwd=version_pkg_path).stdout.strip()
    assert Git().is_immutable_rev_checkout(
        "git+https://g.c/o/r@" + commit, version_pkg_path
    )
    assert not Git().is_immutable_rev_checkout("git+https://g.c/o/r", version_pkg_path)
    assert not Git().is_immutable_rev_checkout(
        "git+https://g.c/o/r@master", version_pkg_path
    )


def test_get_repository_root(script: PipTestEnvironment) -> None:
    version_pkg_path = _create_test_package(script.scratch_path)
    tests_path = version_pkg_path.joinpath("tests")
    tests_path.mkdir()

    root1 = Git.get_repository_root(os.fspath(version_pkg_path))
    assert root1 is not None
    assert os.path.normcase(root1) == os.path.normcase(version_pkg_path)

    root2 = Git.get_repository_root(os.fspath(tests_path))
    assert root2 is not None
    assert os.path.normcase(root2) == os.path.normcase(version_pkg_path)


def test_resolve_commit_not_on_branch(
    script: PipTestEnvironment, tmp_path: pathlib.Path
) -> None:
    repo_path = tmp_path / "repo"
    repo_file = repo_path / "file.txt"
    clone_path = repo_path / "clone"
    repo_path.mkdir()
    script.run("git", "init", cwd=str(repo_path))

    repo_file.write_text(".")
    script.run("git", "add", "file.txt", cwd=str(repo_path))
    script.run("git", "commit", "-m", "initial commit", cwd=str(repo_path))
    script.run("git", "checkout", "-b", "abranch", cwd=str(repo_path))

    # create a commit
    repo_file.write_text("..")
    script.run("git", "commit", "-a", "-m", "commit 1", cwd=str(repo_path))
    commit = script.run("git", "rev-parse", "HEAD", cwd=str(repo_path)).stdout.strip()

    # make sure our commit is not on a branch
    script.run("git", "checkout", "master", cwd=str(repo_path))
    script.run("git", "branch", "-D", "abranch", cwd=str(repo_path))

    # create a ref that points to our commit
    (repo_path / ".git" / "refs" / "myrefs").mkdir(parents=True)
    (repo_path / ".git" / "refs" / "myrefs" / "myref").write_text(commit)

    # check we can fetch our commit
    rev_options = Git.make_rev_options(commit)
    Git().fetch_new(
        str(clone_path),
        HiddenText(repo_path.as_uri(), redacted="*"),
        rev_options,
        verbosity=0,
    )


def _initialize_clonetest_server(
    repo_path: pathlib.Path, script: PipTestEnvironment, enable_partial_clone: bool
) -> pathlib.Path:
    repo_path.mkdir()
    script.run("git", "init", cwd=str(repo_path))
    repo_file = repo_path / "file.txt"
    repo_file.write_text(".")
    script.run("git", "add", "file.txt", cwd=str(repo_path))
    script.run("git", "commit", "-m", "initial commit", cwd=str(repo_path))

    # Enable filtering support on server
    if enable_partial_clone:
        script.run("git", "config", "uploadpack.allowFilter", "true", cwd=repo_path)
        script.run(
            "git", "config", "uploadpack.allowanysha1inwant", "true", cwd=repo_path
        )

    return repo_file


@pytest.mark.parametrize(
    "version_out, expected_message",
    (
        ("git version -2.25.1", "Can't parse git version: git version -2.25.1"),
        ("git version 2.a.1", "Can't parse git version: git version 2.a.1"),
        ("git ver. 2.25.1", "Can't parse git version: git ver. 2.25.1"),
    ),
)
@patch("pip._internal.vcs.versioncontrol.VersionControl.run_command")
def test_git_parse_fail_warning(
    mock_run_command: Mock,
    caplog: pytest.LogCaptureFixture,
    version_out: str,
    expected_message: str,
) -> None:
    """Test invalid git version logs adds an explicit warning log."""
    mock_run_command.return_value = version_out

    caplog.set_level(logging.WARNING)

    git_tuple = Git().get_git_version()
    # Returns an empty tuple if it is an invalid git version
    assert git_tuple == ()

    # Check for warning log
    assert expected_message in caplog.text.strip()


@pytest.mark.skipif(Git().get_git_version() < (2, 17), reason="git too old")
def test_partial_clone(script: PipTestEnvironment, tmp_path: pathlib.Path) -> None:
    """Test partial clone w/ a git-server that supports it"""
    repo_path = tmp_path / "repo"
    repo_file = _initialize_clonetest_server(
        repo_path, script, enable_partial_clone=True
    )
    clone_path1 = repo_path / "clone1"
    clone_path2 = repo_path / "clone2"

    commit = script.run("git", "rev-parse", "HEAD", cwd=str(repo_path)).stdout.strip()

    # Check that we can clone at HEAD
    Git().fetch_new(
        str(clone_path1),
        HiddenText(repo_path.as_uri(), redacted="*"),
        Git.make_rev_options(),
        verbosity=0,
    )
    # Check that we can clone to commit
    Git().fetch_new(
        str(clone_path2),
        HiddenText(repo_path.as_uri(), redacted="*"),
        Git.make_rev_options(commit),
        verbosity=0,
    )

    # Write some additional stuff to git pull
    repo_file.write_text("..")
    script.run("git", "commit", "-am", "second commit", cwd=str(repo_path))

    # Make sure git pull works - with server supporting filtering
    assert (
        "warning: filtering not recognized by server, ignoring"
        not in script.run("git", "pull", cwd=clone_path1).stderr
    )
    assert (
        "warning: filtering not recognized by server, ignoring"
        not in script.run("git", "pull", cwd=clone_path2).stderr
    )


@pytest.mark.skipif(Git().get_git_version() < (2, 17), reason="git too old")
def test_partial_clone_without_server_support(
    script: PipTestEnvironment, tmp_path: pathlib.Path
) -> None:
    """Test partial clone w/ a git-server that does not support it"""
    repo_path = tmp_path / "repo"
    repo_file = _initialize_clonetest_server(
        repo_path, script, enable_partial_clone=False
    )
    clone_path1 = repo_path / "clone1"
    clone_path2 = repo_path / "clone2"

    commit = script.run("git", "rev-parse", "HEAD", cwd=str(repo_path)).stdout.strip()

    # Check that we can clone at HEAD
    Git().fetch_new(
        str(clone_path1),
        HiddenText(repo_path.as_uri(), redacted="*"),
        Git.make_rev_options(),
        verbosity=0,
    )
    # Check that we can clone to commit
    Git().fetch_new(
        str(clone_path2),
        HiddenText(repo_path.as_uri(), redacted="*"),
        Git.make_rev_options(commit),
        verbosity=0,
    )

    # Write some additional stuff to git pull
    repo_file.write_text("..")
    script.run("git", "commit", "-am", "second commit", cwd=str(repo_path))

    # Make sure git pull works - even though server doesn't support filtering
    assert (
        "warning: filtering not recognized by server, ignoring"
        in script.run("git", "pull", cwd=clone_path1).stderr
    )
    assert (
        "warning: filtering not recognized by server, ignoring"
        in script.run("git", "pull", cwd=clone_path2).stderr
    )


def test_clone_without_partial_clone_support(
    script: PipTestEnvironment, tmp_path: pathlib.Path
) -> None:
    """Older git clients don't support partial clone. Test the fallback path"""
    repo_path = tmp_path / "repo"
    repo_file = _initialize_clonetest_server(
        repo_path, script, enable_partial_clone=True
    )
    clone_path = repo_path / "clone1"

    # Check that we can clone w/ old version of git w/o --filter
    with patch("pip._internal.vcs.git.Git.get_git_version", return_value=(2, 16)):
        Git().fetch_new(
            str(clone_path),
            HiddenText(repo_path.as_uri(), redacted="*"),
            Git.make_rev_options(),
            verbosity=0,
        )

    repo_file.write_text("...")
    script.run("git", "commit", "-am", "third commit", cwd=str(repo_path))

    # Should work fine w/o attempting to use `--filter` args
    assert (
        "warning: filtering not recognized by server, ignoring"
        not in script.run("git", "pull", cwd=clone_path).stderr
    )



# -*- coding: utf-8 -*-

import sys
from distutils.core import setup


class FakeError(Exception):
    pass


if sys.argv[1] == "install":
    if hasattr(sys.stdout, "buffer"):
        sys.stdout.buffer.write(
            "\nThis package prints out UTF-8 stuff like:\n".encode("utf-8")
        )
        sys.stdout.buffer.write(
            "* return type of ‘main’ is not ‘int’\n".encode("utf-8")
        )
        sys.stdout.buffer.write(
            "* Björk Guðmundsdóttir [ˈpjœr̥k ˈkvʏðmʏntsˌtoʊhtɪr]".encode("utf-8")
        )
    else:
        pass
        sys.stdout.write("\nThis package prints out UTF-8 stuff like:\n")
        sys.stdout.write(
            "* return type of \xe2\x80\x98main\xe2\x80\x99 is not \xe2\x80\x98int\xe2\x80\x99\n"
        )
        sys.stdout.write(
            "* Bj\xc3\xb6rk Gu\xc3\xb0mundsd\xc3\xb3ttir [\xcb\x88pj\xc5\x93r\xcc\xa5k \xcb\x88kv\xca\x8f\xc3\xb0m\xca\x8fnts\xcb\x8cto\xca\x8aht\xc9\xaar]\n"
        )

    raise FakeError("this package designed to fail on install")

setup(
    name="broken",
    version="0.2",
    py_modules=["broken"],
)



# mypy: no-warn-unused-ignores

import os
import sys
from unittest import mock

import pytest
from pip._vendor import platformdirs

from pip._internal.utils import appdirs


class TestUserCacheDir:
    @pytest.mark.skipif(sys.platform != "win32", reason="Windows-only test")
    def test_user_cache_dir_win(self, monkeypatch: pytest.MonkeyPatch) -> None:
        _get_win_folder = mock.Mock(return_value="C:\\Users\\test\\AppData\\Local")

        monkeypatch.setattr(
            platformdirs.windows,  # type: ignore
            "get_win_folder",
            _get_win_folder,
            raising=False,
        )

        assert (
            appdirs.user_cache_dir("pip")
            == "C:\\Users\\test\\AppData\\Local\\pip\\Cache"
        )
        assert _get_win_folder.call_args_list == [mock.call("CSIDL_LOCAL_APPDATA")]

    @pytest.mark.skipif(sys.platform != "darwin", reason="MacOS-only test")
    def test_user_cache_dir_osx(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("HOME", "/home/test")

        assert appdirs.user_cache_dir("pip") == "/home/test/Library/Caches/pip"

    @pytest.mark.skipif(sys.platform != "linux", reason="Linux-only test")
    def test_user_cache_dir_linux(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.delenv("XDG_CACHE_HOME", raising=False)
        monkeypatch.setenv("HOME", "/home/test")

        assert appdirs.user_cache_dir("pip") == "/home/test/.cache/pip"

    @pytest.mark.skipif(sys.platform != "linux", reason="Linux-only test")
    def test_user_cache_dir_linux_override(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setenv("XDG_CACHE_HOME", "/home/test/.other-cache")
        monkeypatch.setenv("HOME", "/home/test")

        assert appdirs.user_cache_dir("pip") == "/home/test/.other-cache/pip"

    @pytest.mark.skipif(sys.platform != "linux", reason="Linux-only test")
    def test_user_cache_dir_linux_home_slash(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        # Verify that we are not affected by https://bugs.python.org/issue14768
        monkeypatch.delenv("XDG_CACHE_HOME", raising=False)
        monkeypatch.setenv("HOME", "/")

        assert appdirs.user_cache_dir("pip") == "/.cache/pip"

    def test_user_cache_dir_unicode(self, monkeypatch: pytest.MonkeyPatch) -> None:
        if sys.platform != "win32":
            return

        def my_get_win_folder(csidl_name: str) -> str:
            return "\u00DF\u00E4\u03B1\u20AC"

        monkeypatch.setattr(
            platformdirs.windows,  # type: ignore
            "get_win_folder",
            my_get_win_folder,
        )

        # Do not use the isinstance expression directly in the
        # assert statement, as the Unicode characters in the result
        # cause pytest to fail with an internal error on Python 2.7
        result_is_str = isinstance(appdirs.user_cache_dir("test"), str)
        assert result_is_str, "user_cache_dir did not return a str"

        # Test against regression #3463
        from pip._internal.cli.main_parser import create_main_parser

        create_main_parser().print_help()  # This should not crash


class TestSiteConfigDirs:
    @pytest.mark.skipif(sys.platform != "win32", reason="Windows-only test")
    def test_site_config_dirs_win(self, monkeypatch: pytest.MonkeyPatch) -> None:
        _get_win_folder = mock.Mock(return_value="C:\\ProgramData")

        monkeypatch.setattr(
            platformdirs.windows,  # type: ignore
            "get_win_folder",
            _get_win_folder,
            raising=False,
        )

        assert appdirs.site_config_dirs("pip") == ["C:\\ProgramData\\pip"]
        assert _get_win_folder.call_args_list == [mock.call("CSIDL_COMMON_APPDATA")]

    @pytest.mark.skipif(sys.platform != "darwin", reason="MacOS-only test")
    def test_site_config_dirs_osx(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("HOME", "/home/test")

        assert appdirs.site_config_dirs("pip") == [
            "/Library/Application Support/pip",
        ]

    @pytest.mark.skipif(sys.platform != "linux", reason="Linux-only test")
    def test_site_config_dirs_linux(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.delenv("XDG_CONFIG_DIRS", raising=False)

        assert appdirs.site_config_dirs("pip") == [
            "/etc/xdg/pip",
            "/etc",
        ]

    @pytest.mark.skipif(sys.platform != "linux", reason="Linux-only test")
    def test_site_config_dirs_linux_override(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setattr(os, "pathsep", ":")
        monkeypatch.setenv("XDG_CONFIG_DIRS", "/spam:/etc:/etc/xdg")

        assert appdirs.site_config_dirs("pip") == [
            "/spam/pip",
            "/etc/pip",
            "/etc/xdg/pip",
            "/etc",
        ]

    @pytest.mark.skipif(sys.platform != "linux", reason="Linux-only test")
    def test_site_config_dirs_linux_empty(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setattr(os, "pathsep", ":")
        monkeypatch.setenv("XDG_CONFIG_DIRS", "")
        assert appdirs.site_config_dirs("pip") == [
            "/etc/xdg/pip",
            "/etc",
        ]


class TestUserConfigDir:
    @pytest.mark.skipif(sys.platform != "win32", reason="Windows-only test")
    def test_user_config_dir_win_no_roaming(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        _get_win_folder = mock.Mock(return_value="C:\\Users\\test\\AppData\\Local")

        monkeypatch.setattr(
            platformdirs.windows,  # type: ignore
            "get_win_folder",
            _get_win_folder,
            raising=False,
        )

        assert (
            appdirs.user_config_dir("pip", roaming=False)
            == "C:\\Users\\test\\AppData\\Local\\pip"
        )
        assert _get_win_folder.call_args_list == [mock.call("CSIDL_LOCAL_APPDATA")]

    @pytest.mark.skipif(sys.platform != "win32", reason="Windows-only test")
    def test_user_config_dir_win_yes_roaming(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        _get_win_folder = mock.Mock(return_value="C:\\Users\\test\\AppData\\Roaming")

        monkeypatch.setattr(
            platformdirs.windows,  # type: ignore
            "get_win_folder",
            _get_win_folder,
            raising=False,
        )

        assert (
            appdirs.user_config_dir("pip") == "C:\\Users\\test\\AppData\\Roaming\\pip"
        )
        assert _get_win_folder.call_args_list == [mock.call("CSIDL_APPDATA")]

    @pytest.mark.skipif(sys.platform != "darwin", reason="MacOS-only test")
    def test_user_config_dir_osx(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("HOME", "/home/test")

        if os.path.isdir("/home/test/Library/Application Support/"):
            assert (
                appdirs.user_config_dir("pip")
                == "/home/test/Library/Application Support/pip"
            )
        else:
            assert appdirs.user_config_dir("pip") == "/home/test/.config/pip"

    @pytest.mark.skipif(sys.platform != "linux", reason="Linux-only test")
    def test_user_config_dir_linux(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
        monkeypatch.setenv("HOME", "/home/test")

        assert appdirs.user_config_dir("pip") == "/home/test/.config/pip"

    @pytest.mark.skipif(sys.platform != "linux", reason="Linux-only test")
    def test_user_config_dir_linux_override(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        monkeypatch.setenv("XDG_CONFIG_HOME", "/home/test/.other-config")
        monkeypatch.setenv("HOME", "/home/test")

        assert appdirs.user_config_dir("pip") == "/home/test/.other-config/pip"

    @pytest.mark.skipif(sys.platform != "linux", reason="Linux-only test")
    def test_user_config_dir_linux_home_slash(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        # Verify that we are not affected by https://bugs.python.org/issue14768
        monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
        monkeypatch.setenv("HOME", "/")

        assert appdirs.user_config_dir("pip") == "/.config/pip"



import os
import subprocess
from pathlib import Path
from typing import List, Tuple

_BROKEN_STDOUT_RETURN_CODE = 120


def setup_broken_stdout_test(
    args: List[str], deprecated_python: bool
) -> Tuple[str, int]:
    proc = subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    # Call close() on stdout to cause a broken pipe.
    assert proc.stdout is not None
    proc.stdout.close()
    returncode = proc.wait()
    assert proc.stderr is not None
    stderr = proc.stderr.read().decode("utf-8")

    expected_msg = "ERROR: Pipe to stdout was broken"
    if deprecated_python:
        assert expected_msg in stderr
    else:
        assert stderr.startswith(expected_msg)

    return stderr, returncode


def test_broken_stdout_pipe(deprecated_python: bool) -> None:
    """
    Test a broken pipe to stdout.
    """
    stderr, returncode = setup_broken_stdout_test(
        ["pip", "list"],
        deprecated_python=deprecated_python,
    )

    # Check that no traceback occurs.
    assert "raise BrokenStdoutLoggingError()" not in stderr
    assert stderr.count("Traceback") == 0

    assert returncode == _BROKEN_STDOUT_RETURN_CODE


def test_broken_stdout_pipe__log_option(deprecated_python: bool, tmpdir: Path) -> None:
    """
    Test a broken pipe to stdout when --log is passed.
    """
    log_path = os.path.join(str(tmpdir), "log.txt")
    stderr, returncode = setup_broken_stdout_test(
        ["pip", "--log", log_path, "list"],
        deprecated_python=deprecated_python,
    )

    # Check that no traceback occurs.
    assert "raise BrokenStdoutLoggingError()" not in stderr
    assert stderr.count("Traceback") == 0

    assert returncode == _BROKEN_STDOUT_RETURN_CODE


def test_broken_stdout_pipe__verbose(deprecated_python: bool) -> None:
    """
    Test a broken pipe to stdout with verbose logging enabled.
    """
    stderr, returncode = setup_broken_stdout_test(
        ["pip", "-vv", "list"],
        deprecated_python=deprecated_python,
    )

    # Check that a traceback occurs and that it occurs at most once.
    # We permit up to two because the exception can be chained.
    assert "raise BrokenStdoutLoggingError()" in stderr
    assert 1 <= stderr.count("Traceback") <= 2

    assert returncode == _BROKEN_STDOUT_RETURN_CODE



import os

from setuptools import find_packages, setup


def path_to_url(path):
    """
    Convert a path to URI. The path will be made absolute and
    will not have quoted path parts.
    """
    path = os.path.normpath(os.path.abspath(path))
    drive, path = os.path.splitdrive(path)
    filepath = path.split(os.path.sep)
    url = "/".join(filepath)
    if drive:
        return "file:///" + drive + url
    return "file://" + url


setup(
    name="LocalEnvironMarker",
    version="0.0.1",
    packages=find_packages(),
    extras_require={
        ":python_version == '2.7'": ["simple"],
    },
)
